
################################################################################
#################                    AAIndex                   #################
################################################################################

#importing required modules and dependencies
import json
import numpy as np
import pandas as pd
import os
import matplotlib.pyplot as plt
import numpy as np
from sklearn.preprocessing import OneHotEncoder, LabelEncoder
import sys, copy, re
import requests
import shutil
import csv
import urllib.request as request
from contextlib import closing
from collections import defaultdict
import itertools
from difflib import get_close_matches
from .globals_ import DATA_DIR, OUTPUT_DIR, OUTPUT_FOLDER

#look into and remove ('-') from get_encoding & get_amino_acids functions.
#fix     def get_record_from_name(self, name): <- not returning anything
class AAIndex():
    """
            Python parser for AAindex1: Amino Acid Index Database
                      (**abbreviated to AAI onwards**)

    The AAindex is a database of numerical indices representing various physicochemical
    and biochemical properties of amino acids and pairs of amino acids. The focus
    on this class is on the AAIndex 1 database which stores the amino acid index of
    20 numerical values for the 20 amino acids. http://www.genome.jp/aaindex/

    Some of the parsing functionality is inspired and based off AAIndex1 parser
    aaindex2json.py by harmslab:
    https://github.com/harmslab/hops/blob/master/hops/features/data/util/aaindex2json.py

    Data format of AAI1:
    ************************************************************************
    *                                                                      *
    * Each entry has the following format.                                 *
    *                                                                      *
    * H Accession number                                                   *
    * D Data description                                                   *
    * R PMID                                                               *
    * A Author(s)                                                          *
    * T Title of the article                                               *
    * J Journal reference                                                  *
    * * Comment or missing                                                 *
    * C Accession numbers of similar entries with the correlation          *
    *   coefficients of 0.8 (-0.8) or more (less).                         *
    *   Notice: The correlation coefficient is calculated with zeros       *
    *   filled for missing values.                                         *
    * I Amino acid index data in the following order                       *
    *   Ala    Arg    Asn    Asp    Cys    Gln    Glu    Gly    His    Ile *
    *   Leu    Lys    Met    Phe    Pro    Ser    Thr    Trp    Tyr    Val *
    * //                                                                   *
    ************************************************************************
    --------------------------------------------------------------------------------
    H ANDN920101
    D alpha-CH chemical shifts (Andersen et al., 1992)
    R PMID:1575719
    A Andersen, N.H., Cao, B. and Chen, C.
    T Peptide/protein structure analysis using the chemical shift index method:
      upfield alpha-CH values reveal dynamic helices and aL sites
    J Biochem. and Biophys. Res. Comm. 184, 1008-1014 (1992)
    C BUNA790102    0.949
    I    A/L     R/K     N/M     D/F     C/P     Q/S     E/T     G/W     H/Y     I/V
        4.35    4.38    4.75    4.76    4.65    4.37    4.29    3.97    4.63    3.95
        4.17    4.36    4.52    4.66    4.44    4.50    4.35    4.70    4.60    3.95
    --------------------------------------------------------------------------------

    Attributes
    ----------
    aaindex_filename : str (default = 'aaindex1')
        local filename of aaindex1 file that class will import the database
        from in the data directory, default value of 'aaindex1' is reccomended.
    download_using : str (default = "ftp")
        decide to download AAI database using ftp or https, ftp used by default.

    Methods
    -------
    parse_aaindex_to_json():
    parse_aaindex_to_category(aaindex_category_file):
    download_aaindex():
    get_amino_acids():
    get_amino_acids_encoding():
    get_record_codes():
    get_num_records():
    get_record_names():
    get_record_from_code(code):
    get_record_from_name():
    get_values_from_record():
    get_ref_from_record():
    get_category_from_record():

    References
    ----------
    [1] S. Kawashima and M. Kanehisa, “AAindex: amino acid index database,”
        Nucleic Acids Res., vol. 28, no. 1, p. 374, 2000.
    [2] L. C. Wheeler, A. Perkins, C. E. Wong, and M. J. Harms, “Learning peptide
        recognition rules for a low-specificity protein,” Protein Sci., vol. 29,
        no. 11, pp. 2259–2273, 2020.
    """
    def __init__(self, aaindex_filename='aaindex1', download_using='ftp'):

        self.aaindex_filename = aaindex_filename
        self.download_using = download_using

        #download AAI database using ftp or https
        if self.download_using=='ftp':
            self.url = "ftp://ftp.genome.jp/pub/db/community/aaindex/aaindex1"
        elif self.download_using=='http' or download_using=='https':
            self.url = "https://www.genome.jp/ftp/db/community/aaindex/aaindex1"
        else:
            self.url = "ftp://ftp.genome.jp/pub/db/community/aaindex/aaindex1"

        #if AAIndex database not found in data directory then call download method
        if not (os.path.isfile(os.path.join(DATA_DIR,self.aaindex_filename+'.json'))):
            if not (os.path.isfile(os.path.join(DATA_DIR,self.aaindex_filename))):
                self.download_aaindex()

        #if parsed json of AAIndex already in file then read it and return  <- Not working at the moment
        # if (os.path.isfile(os.path.join(DATA_DIR, self.aaindex_filename +'.json'))):
        #     self.aaindex_json = json.load(os.path.join(DATA_DIR, self.aaindex_filename +'.json'))
        # else:
        #parse AAIndex database file into JSON format
        self.aaindex_json = self.parse_aaindex_to_json()

        self.categories = None

    def parse_aaindex_to_json(self):
        """
        Parse AAI database into JSON format. Each AAI record will be indexed by
        its feature code/index code, and will be in the format as shown in the
        example above. The file will be stored in a json file called according
        to (self.aaindex_filename+'.json') variable.

        Returns
        -------
        aaindex_json : dict
          parsed AAI database in dict form.
        """
        #initialise keys of AAI database
        template_dict = {"H":[],
                         "D":[],
                         "R":[],
                         "A":[],
                         "*":[],
                         "T":[],
                         "J":[],
                         "C":[],
                         "I":[]}

        #open AAI file for reading and parsing, by default it should be stored in DATA_DIR
        try:
            tmp_filepath = os.path.join(DATA_DIR,self.aaindex_filename)
            f = open(tmp_filepath,'r')
        except IOError:
            print('Error opening file, check filename = {} and is stored in \
                    {} directory.'.format(self.aaindex_filename, DATA_DIR))

        #read lines of file
        lines = f.readlines()
        f.close()

        clean_up_pattern = re.compile("\"")

        #initilaise parsed AAI database dictionary
        aaindex_json = {}
        current_dict = copy.deepcopy(template_dict)

        '''
        iterate through each line in the AAI database, parsing each record and its
        amino acid values into its own entry in the dictionary. Each index/record
        is seperated by a '//'. Remove any duplicate records, set any missing
        ('-') or NA amino acid values to 0. Store resulting dict into aaindex_json
        instance variable.
        '''
        for l in lines:

            if l.startswith("//"):

                # deal with meta data
                name = " ".join(current_dict["H"])
                name = clean_up_pattern.sub("'",name)

                description = " ".join(current_dict["D"])
                description = clean_up_pattern.sub("'",description)

                citation = "{} '{}' {}".format(" ".join(current_dict["A"]),
                                               " ".join(current_dict["T"]),
                                               " ".join(current_dict["J"]))

                citation = citation + "; Kawashima, S. and Kanehisa, M. \
                    'AAindex: amino acid index database.'  Nucleic Acids Res. 28, 374 (2000)."

                citation = clean_up_pattern.sub("'",citation)

                notes = " ".join(current_dict["*"])
                notes = clean_up_pattern.sub("'",notes)

                # parse amino acid data
                aa_lines = current_dict["I"]

                aa_names = aa_lines[0].split()
                row_0_names = [aa.split("/")[0] for aa in aa_names]
                row_1_names = [aa.split("/")[1] for aa in aa_names]

                row_0_values = aa_lines[1].split()
                row_1_values = aa_lines[2].split()

                values = {}
                for i in range(len(row_0_values)):
                    try:
                        values[row_0_names[i]] = float(row_0_values[i])
                    except ValueError:
                        values[row_0_names[i]] = "NA"

                    try:
                        values[row_1_names[i]] = float(row_1_values[i])
                    except ValueError:
                        values[row_1_names[i]] = "NA"

                # look for duplicate name entries
                try:
                    aaindex_json[name]
                    err = "duplicate value name ({})".format(name)
                    raise ValueError('Duplicate AAI Record name found')
                except KeyError:
                    pass

                aaindex_json[name] = {"description":description,
                                  "refs":citation,
                                  "notes":notes,
                                  "values":values}

                current_dict = copy.deepcopy(template_dict)
                continue

            this_entry = l[0]
            if l[0] != " ":
                current_entry = this_entry

            current_dict[current_entry].append(l[1:].strip())

        #append '-' to each aa index entry to account for missing AA in a protein sequence
        for index in aaindex_json:
            aaindex_json[index]['values']['-'] = 0

            #set any NA amino acid values to 0
            for val in aaindex_json[index]['values']:
                if aaindex_json[index]['values'][val] == 'NA':
                    aaindex_json[index]['values'][val] = 0

        #save parsed dictionary into JSON format to DATA_DIR
        with open((os.path.join(DATA_DIR, self.aaindex_filename + '.json')),'w') as output_F:
          json.dump(aaindex_json, output_F,indent=4, sort_keys=True)

        return aaindex_json

    def download_aaindex(self, save_dir=DATA_DIR):
        """
        If AAI database not found in DATA directory, then it will be downloaded from
        the dedicated FTP or HTTPS server from https://www.genome.jp/aaindex/.
        FTP is the default method used for downloading the database with the URL:
        "ftp://ftp.genome.jp/pub/db/community/aaindex/aaindex1". If you want to
        download by https, set the 'download_using' instance variable to 'https'.

        Parameters
        ----------
        save_dir : str (default = DATA_DIR)
            Directory to save the AAI database to. By default it wil be stored in
            the global var DATA_DIR = 'data'.
        """
        #if directory doesnt exist then create it
        if not os.path.isdir(save_dir):
            try:
                os.makedirs(save_dir)
            except:
                raise OSError('Error creating save directory: {}'.format(save_dir))

        #fetch AAI database from URL if not present in DATA_DIR
        try:
            if not(os.path.isfile(os.path.join(save_dir, self.aaindex_filename))):
                try:
                    with closing(request.urlopen(self.url)) as r:
                        with open((os.path.join(save_dir, self.aaindex_filename)), 'wb') as f:
                            shutil.copyfileobj(r, f)
                    print('AAIndex1 successfully downloaded.')
                except requests.exceptions.RequestException:
                    print('Error downloading or exporting AAIndex from the url {}.'.format(self.url))
            else:
                pass    #AAIndex already in folder
                print('AAIndex already exists in folder.')
        except:
            raise OSError('Save Directory does not exist: {}.'.format(save_dir))

    def parse_aaindex_to_category(self, aaindex_category_file='aaindex-to-category.txt'):
        """
        Parse category file which maps each AAI record in the databse into 1 of 8 categories.
        Category file and parsing code inspired from:
        https://github.com/harmslab/hops

        Parameters
        ----------
        aaindex_category_file : str
            Name of category file to parse (default is "aaindex-to-category.txt").

        Returns
        -------
        aaindex_category : dict
            Dictionary that maps each AAI record into 1 of 8 categories.
        """
        #if input parameter is a full path, read it else read from default DATA_DIR
        if os.path.isfile(aaindex_category_file):
            f = open(aaindex_category_file,'r')
        else:
            try:
                f = open((os.path.join(DATA_DIR, aaindex_category_file)),'r')
            except IOError:
                print('Error opening AAIndex1 category file, check {} in {} directory'
                    .format(aaindex_category_file, DATA_DIR))

        #get total number of lines in file
        total_lines = len(f.readlines(  ))

        #open new file in data directory to store parsed category file
        f.seek(0)
        category_output_file = "aaindex-to-category-parsed.txt"
        f_out = open((os.path.join(DATA_DIR, category_output_file)), "w")

        #lines starting with '#' are file metadata so don't write these to parsed output file
        for line in f.readlines():
            if not (line.startswith('#')):
                f_out.write(line)
        f.close()
        f_out.close()

        #open parsed category file for reading
        with open(os.path.join(DATA_DIR, category_output_file)) as f_out:
            reader = csv.reader(f_out, delimiter="\t")
            d = list(reader)
        f_out.close()

        aaindex_category = {}

        #iterate through all lines in parsed file and store AAI indices as keys
        #   in the output dict and their respective categories as the values of dict
        for i in range(0,len(d)):
          for j in range(0,len(d[i])):
            category_substring = d[i][1].strip()
            category_substring = category_substring.split(" ",1)
            aaindex_category[d[i][0]] = category_substring[0]

        self.categories = aaindex_category

    def get_amino_acids(self):
        """
        Get all canonical amino acid letters. The '-' value will also be included
        in the list from this function as it accounts for the abcense of any
        amino acid or gaps in a AAI record.

        Returns
        -------
        amino_acids : list
            List of all 20 canoniocal amino acid letters as found in each record
            of the AAI database.
        """
        amino_acids = list(self.aaindex_json[list(self.aaindex_json.keys())[0]]["values"].keys())
        amino_acids.sort()

        return amino_acids

    def get_amino_acids_encoding(self):
      """
      Get one-hot encoding of amino acids.

      Returns
      -------
      onehot_encoded : np.ndarray
        one hot encoded array of the 20 canonical amino acids.
      """
      all_amino_acids = self.get_amino_acids()
      values = np.array(all_amino_acids[1:])    #convert amino acids to np array

      #encode amino acids with value between 0 and n_classes-1.
      label_encoder = LabelEncoder()
      integer_encoded = label_encoder.fit_transform(values)

      #encode amino acids as a one-hot numeric array.
      onehot_encoder = OneHotEncoder(sparse=False)
      integer_encoded = integer_encoded.reshape(len(integer_encoded), 1)
      onehot_encoded = onehot_encoder.fit_transform(integer_encoded)

      return onehot_encoded

    def get_record_codes(self):
      """
      Get list of all AAI index codes for each record in the database.

      Returns
      -------
      records : list
        list of record names/index codes for all records in the AAI database.
      """
      records = list(self.aaindex_json.keys())
      records.sort()       #sort into alphabetical order

      return records

    def get_num_records(self):
      """
      Calculate total number of records/indices in the AAI database.

      Returns
      -------
      len(self.get_record_codes()) : int
        number of indices/records found in the AAI database.
      """
      return len(self.get_record_codes())

    # def get_feature_names(self):
    def get_record_names(self):
      """
      Return a list of all index descriptions for all records in the AAI database.

      Returns
      -------
      desc : list
         list of descriptions for all records in the AAI database.
      """
      desc = []

      #iterate through database, appending all descriptions to desc list
      for name in list(self.aaindex_json.values()):
        desc.append(name['description'])

      return desc

    def get_record_from_code(self, index_code):
      """
      Return full AAI database record details from its feature/index code.

      Parameters
      ----------
      index_code : str
         AAI database record feature code/index code.

      Returns
      -------
      record : dict
        dict of AAI database record.
      """
      #stripping input of whitespace
      try:
          index_code = index_code.strip()
      except:
        raise TypeError('Input parameter {} is not of correct datatype string, got {}' \
            .format(index_code, type(index_code)))

      #check that inputted index_code does exist in the AAI database
      if index_code not in (self.get_record_codes()):
          raise ValueError('Record Index ({}) not found in AAIndex'.format(index_code))

      record = (self.aaindex_json[index_code])

      return record

    def get_record_from_name(self, name):
        """
        Return full AAI database record details from its description. Search
        through the descriptions/names of all records in the database, returning
        record that matches name input parameter.

        Parameters
        ----------
        name : str
            AAI database record feature name/description.

        Returns
        -------
        feature : dict/None
            dict of full AAI database record or None if not found.
        """
        correct_index = 0
        for index, value in self.aaindex_json.items():
          if value['description'] != name:
           continue
          else:
           correct_index = index
           return self.aaindex_json[correct_index]

        return None

    def get_values_from_record(self, index_code):
      """
      Return amino acid values from database record of index in the AAI from
      its feature/index code.

      Parameters
      ----------
      index_code : str
          AAI database record feature code/index code.

      Returns
      -------
      values : dict
        amino acid values for specified AAI database record.
      """
      #stripping input of whitespace
      try:
          index_code.strip()
      except:
        raise TypeError('Input parameter {} is not of correct datatype string, got {}' \
            .format(index_code, type(index_code)))

      #check that inputted index_code does exist in the AAI database
      if index_code not in (self.get_record_codes()):
            raise ValueError('Record index {} not found in AAI'.format(index_code))

      #get amino acid values for specified index
      values = (self.aaindex_json[index_code]['values'])

      return values

    def get_ref_from_record(self, index_code):
      """
      Return reference details of a AAI database record from its feaure/index code.

      Parameters
      ----------
      index_code : str
        AAI database record feature code/index code.

      Returns
      -------
      refs : list
        reference details for specifed AAI database record.
      """
      #stripping input of whitespace
      try:
          index_code.strip()
      except:
        raise TypeError('Input parameter {} is not of correct datatype string, got {}' \
            .format(index_code, type(index_code)))

      #check that inputted index_code does exist in the AAI database
      if index_code not in (self.get_record_codes()):
            raise ValueError('Record index {} not found in AAI'.format(index_code))

      refs = (self.aaindex_json[index_code]['refs'])    #get refs from record

      return refs

    def get_category_from_record(self, index_code):
      """
      Return category of a AAI database record from its feaure/index code.

      Parameters
      ----------
      index_code : str
            AAI database record feature code/index code.

      Returns
      -------
      cat : str
        category of AAI database record according to parsed aaindex-to-category.txt file.
      """
      #stripping input of whitespace
      try:
          index_code.strip()
      except:
        raise TypeError('Input parameter {} is not of correct datatype string, got {}' \
            .format(index_code, type(index_code)))

      #check that inputted index_code does exist in the AAI database
      if index_code not in (self.get_record_codes()):
          raise ValueError('Record index {} not in AAI database'.format(index_code))

      if self.categories == None:
          self.parse_aaindex_to_category()

      cat = self.categories[index_code]

      return cat

######################          Getters & Setters          ######################

    @property
    def url(self):
        return self._url

    @url.setter
    def url(self, value):
        self._url = value

    @property
    def aaindex_filename(self):
        return self._aaindex_filename

    @aaindex_filename.setter
    def aaindex_filename(self, value):
        self._aaindex_filename = value

    @property
    def categories(self):
        return self._categories

    @categories.setter
    def categories(self, value):
        self._categories = value

################################################################################

    def __str__(self):
        return "AAIndex1 Database of size {} - stored in {} ".format(
        self.get_num_records(),self.aaindex_filename
        )

    def __repr__(self):
        return (self.aaindex_json)

    def __sizeof__(self):
        """return size of AAI database file"""
        return os.path.getsize(os.path.isfile(os.path.join(DATA_DIR,self.aaindex_filename)))



# -*- coding: utf-8 -*-
#  Copyright (c) 2016-2017, Zhijiang Yao, Jie Dong and Dongsheng Cao
#  All rights reserved.
#  This file is part of the PyBioMed.
#  The contents are covered by the terms of the BSD license
#  which is included in the file license.txt, found at the root
#  of the PyBioMed source tree.
"""
################################################################################################

This module is used to download the protein sequence from the uniprot (http://www.uniprot.org/)

website. You can only need input a protein ID or prepare a file (ID.txt) related to ID. You can

 obtain a .txt (ProteinSequence.txt) file saving protein sequence you need.  You can freely use

 and distribute it. If you hava  any problem, you could contact with us timely!

Authors: Zhijiang Yao and Dongsheng Cao.

Date: 2016.06.04

Email: gadsby@163.com

################################################################################################
"""

try:
    # Python 3
    from urllib.request import urlopen
except ImportError:
    # Python 2
    from urllib2 import urlopen
# Core Library modules
import string


##################################################################################################
def GetProteinSequence(ProteinID):
    """
    #########################################################################################
    Get the protein sequence from the uniprot website by ID.

    Usage:

    result=GetProteinSequence(ProteinID)

    Input: ProteinID is a string indicating ID such as "P48039".

    Output: result is a protein sequence.
    #########################################################################################
    """

    ID = str(ProteinID)
    localfile = urlopen("http://www.uniprot.org/uniprot/" + ID + ".fasta")
    temp = localfile.readlines()
    res = ""
    for i in range(1, len(temp)):
        res = res + string.strip(temp[i])
    return res


##################################################################################################
def GetProteinSequenceFromTxt(path, openfile, savefile):
    """
    #########################################################################################
    Get the protein sequence from the uniprot website by the file containing ID.

    Usage:

    result=GetProteinSequenceFromTxt(path,openfile,savefile)

    Input: path is a directory path containing the ID file such as "/home/orient/protein/"

    openfile is the ID file such as "proteinID.txt"

    savefile is the file saving the obtained protein sequences such as "protein.txt"
    #########################################################################################
    """
    f1 = open(path + savefile, "wb")
    f2 = open(path + openfile, "r")
    # 	res=[]
    for index, i in enumerate(f2):

        itrim = string.strip(i)
        if itrim == "":
            continue
        else:
            temp = GetProteinSequence(itrim)
            print("--------------------------------------------------------")
            print("The %d protein sequence has been downloaded!" % (index + 1))
            print(temp)
            f1.write(temp + "\n")
            print("--------------------------------------------------------")
        # 		res.append(temp+'\n')
        # 	f1.writelines(res)
    f2.close()
    f1.close()
    return 0


##################################################################################################
if __name__ == "__main__":

    localfile = ["P48039"]
    for index, i in enumerate(localfile):
        itrim = string.strip(i)
        if itrim == "":
            continue
        else:
            temp = GetProteinSequence(itrim)
            print("--------------------------------------------------------")
            print("The %d protein sequence has been downloaded!" % (index + 1))
            print(temp)
            print("--------------------------------------------------------")



################################################################################
#################                     pySAR                    #################
################################################################################

import pandas as pd
import numpy as np
import datetime, time
import argparse
import itertools
import pickle
import io
import os
from difflib import get_close_matches
import json

from .globals_ import OUTPUT_DIR, OUTPUT_FOLDER, DATA_DIR
from .aaindex import AAIndex
from .model import Model
from .proDSP import ProDSP
from .evaluate import Evaluate
from .utils import *
from .plots import plot_reg
from .descriptors import Descriptors

class PySAR():
    """
    The PySAR class is the main class for the pySAR software. The class allows for
    the encoding of protein sequences via a plethora of techniques, mainly via
    AAI Indices and or strucutrual and physiochemical protein descriptors that
    are then used as features in the building of a predictive model created to
    map the protein sequences to a sought-after activity value (activity attribute),
    this is known as a sequence activity relationship (SAR). Three main encoding
    strategies are possible in the class, namely using AAI Indices or protein
    descriptors as well as AAI Indices + Descriptors. The class accepts strings
    or lists of AAI Indices or descriptors and then passes these through a pipeline
    to get the required encoding of the respective sequences. The calculated encodings
    of the sequences are used as features in the building of the predictive model
    that will predict the acitivty values for new unseen protein sequences. After
    the encoding process, various metrics will be captured and stored in a local
    output folder according to the OUTPUT_FOLDER global var as well as a regression
    plot showing how well the model fits to the test data.

    Attributes
    ----------
    dataset : str (default = "")
        full path to dataset or name of dataset if it is stored in DATA_DIR.
    seq_col : str (default = "sequence")
        name of column in dataset that stores the protein sequences. By default
        the class will look for a column called 'sequence'.
    activity : str (default = "")
        name of activity column in dataset.
    algorithm : str (default = "")
        name of regression model to use for building the predictive models, class
        will accept full name or approximate name of model e.g "PLSReg", "plsregg" and
        "PLSRegression" will all build a PlsRegression model.
    parameters : dict (default = {})
        dictionary of parameters to use for the predictive model. By default the
        default parameters of the model will be used.
    test_split : float (default = 0.2)
        specifies the proportion of the dataset to use for testing. By default a
        80:20 split will be used, meaning 80% of the data will be used for training
        and 20% for testing.
    descriptors_csv : str (default = "descriptors.csv")
        csv file storing the pre-calculated descriptor values for the sequences
        in the dataset. By default the class will look for a file named
        "descriptors.csv" in the DATA_DIR and will use its contents as the
        descriptor features, instead of having to recalculate all descriptors for the dataset.

    Methods
    -------
    read_data():

    preprocessing()

    get_aai_encoding()

    aai_encoding()

    desc_encoding()

    aai_desc_encoding()

    output_results()
    """
    def __init__(self,dataset="",seq_col="sequence", activity="",algorithm="",\
        parameters={},test_split=0.2, descriptors_csv="descriptors.csv"):

        self.dataset = dataset
        self.seq_col = seq_col
        self.activity = activity
        self.algorithm = algorithm
        self.parameters = parameters
        self.test_split = test_split
        self.descriptors_csv = descriptors_csv
        self.aai_indices = None
        self.descriptors = None

        #create instance of AAIndex class
        self.aaindex = AAIndex()

        #import and read dataset
        self.data = self.read_data()

        #pre-process dataset and protein sequences
        self.preprocessing()                            #Reorder these??

        #get number of rows and cols of dataset
        self.num_seqs = len(self.data[self.seq_col])
        self.seq_len = len(max(self.data[self.seq_col],key=len))

        #create instance of model class of type specified by algorithm
        self.model = Model(algorithm=self.algorithm, parameters=self.parameters)

        #updating algoritm attribute
        self.algoritm = repr(self.model)

        #create output folder to store all model assets and results
        create_output_dir()

    def read_data(self):
        """
        Read in dataset according to name from 'dataset' attribute. By default
        the dataset should be stored in DATA_DIR.

        """
        dataset_file = os.path.join(DATA_DIR,self.dataset)

        #read in dataset csv if found in path, if not raise error
        if not (os.path.isfile(dataset_file)):
            if not (os.path.isfile(self.dataset)):
                raise OSError('Dataset filepath is not correct: {}'.format(self.dataset))
            dataset_file = self.dataset

        try:
            data = pd.read_csv(dataset_file, sep=",", header=0)
            return data
        except:
            raise IOError('Error opening dataset file: {}'.format(dataset_file))

    def preprocessing(self):
        """
        Pre-process protein sequences in dataset. Validate column names, check
        for invalid amino acids in sequences and remove any NAN or +/- infinity values.

        """
        #get closest match for sequence column name in dataset
        seq_col_matches = get_close_matches(self.seq_col,self.data.columns, cutoff=0.4)

        #set sequence col to the first match found, else raise error
        if seq_col_matches!=[]:
            self.seq_col = seq_col_matches[0]
        else:
            raise ValueError('Sequence Column ({}) not present in dataset columns \
                - {} /n '.format(self.seq_col, self.data.columns))

        #remove any gaps found in sequences in dataset
        self.data[self.seq_col] = remove_gaps(self.data[self.seq_col])

        #verify no invalid amino acids found in sequences, if so then raise error
        invalid_seqs = valid_sequence(self.get_seqs())
        if invalid_seqs!=None:
            raise ValueError('Invalid Amino Acids found in protein sequence \
                dataset: {}'.format(invalid_seqs))

        #get closest match for activity column name in dataset
        activity_matches = get_close_matches(self.activity,self.data.columns,cutoff=0.4)

        #set activity col to the first match found, else raise error
        if activity_matches!=[]:
            self.activity = activity_matches[0]
        else:
            raise ValueError('Activity Column ({}) not present in dataset columns  \
                - {} /n '.format(self.activity,list(self.data.columns)))

        #remove any +/- infinity values or any Null/NAN's from activity values
        self.data[self.activity].replace([np.inf,-np.inf], np.nan)
        self.data[self.activity].fillna(0,inplace=True)

    def get_aai_enoding(self, indices=None):
        """
        Get AAI index encoding values for index specified by indices for each amino
        acid in each of the protein sequences in dataset. If multiple indices
        specified then encode protein sequences with each index and concatenate.

        Parameters
        ----------
        indices : str/list (default = None)
            string or list of AAI indices.

        Returns
        -------
        encoded_seqs : np.ndarray
            array of the encoded protein sequences in dataset.
        """
        #validate AAI indices are present in the input parameter, if not raise error
        if indices == None or indices == "":
            raise ValueError('AAI indices input parameter cannot be None or empty.')

        encoded_indices = []

        #if single index input passed into parameter
        if not isinstance(indices, list):

            #get dict of amino acid encoding values for index specified by indices
            encoded_aai = self.aaindex.get_record_from_code(indices)['values']

            #initialise temp arrays to store encoded sequences
            temp_seq_vals = []
            temp_all_seqs = []

            #iterate through each protein sequence and amino acid, getting the AAI index encoding value
            for protein in range(0, len(self.data[self.seq_col])):
                for aa in self.data[self.seq_col][protein]:
                    temp_seq_vals.append(encoded_aai[aa])

                #append encoding and reset temp array
                temp_all_seqs.append(temp_seq_vals)
                temp_seq_vals = []

            #zero-pad encoding list so that sequences are all the same length
            temp_all_seqs = zero_padding(temp_all_seqs)

            #convert list of lists into array
            temp_all_seqs = np.array(temp_all_seqs, dtype="float32")

            encoded_seqs = temp_all_seqs

            return encoded_seqs

        #if multiple indices (list) passed into input parameter
        else:

            #create zeros numpy array to store encoded sequence output
            encoded_aai_ = np.zeros((self.num_seqs, self.seq_len))

            #if multiple indices used then calcualte AAI index encoding for each and
            #   then concatenate after each calculation
            for ind in range(0,len(indices)):
                encoded_aai = self.aaindex.get_record_from_code(indices[ind])['values']

                temp_seq_vals = []
                temp_all_seqs = []

                for protein in range(0, len(self.data[self.seq_col])):
                    for aa in self.data[self.seq_col][protein]:
                        temp_seq_vals.append(encoded_aai[aa])

                    temp_all_seqs.append(temp_seq_vals)
                    temp_seq_vals = []

                temp_all_seqs = zero_padding(temp_all_seqs)

                temp_all_seqs =np.array(temp_all_seqs, dtype="float32")

                #in first iteration through indices set encoded_aai_ to zeros
                #   initialised np array, else concatenate to array in previous iteration
                if ind == 0:
                    encoded_aai_ = temp_all_seqs
                else:
                    encoded_aai_ = np.concatenate((encoded_aai_,temp_all_seqs), axis=1)

            encoded_seqs = encoded_aai_

            return encoded_seqs

    def encode_aai(self,spectrum=None, indices=None, window=None, filter_=None,use_dsp=True):
        """
        Encode using AAI indices. If multiple indices input then calculate each
        and concatenate them. Build predictive model from AAI feature data. The
        resulting model assets and its results will be exported to the directory
        pointed to by the global variable OUTPUT_DIR. If use_dsp is true then
        pass AAI Indices through a DSP transformation specified by DSP parameters:
        spectrum, window & filter.

        Parameters
        ----------
        spectrum : str
            type of protein spectrum to use for DSP transformation.
        indices : str/list (default = None)
            string or list of indices from the AAI.
        window : str
            name of window function to apply to Fourier Transform.
        filter_ : str
            name of filter function to apply to ouput of Fourier Transform.
        use_dsp : bool (default = True)
            whether to transform AAI index encodings via DSP techniques and a Fourier
            transform. If false then no protein spectra will be generated.

        Returns
        -------
        aai_df : pd.Series
            pandas series storing metrics and results of encoding.
        """
        #validate AAI indices are present in the input parameter
        if indices == None or indices == "":
            raise ValueError('AAI indices input parameter cannot be None or empty.')

        self.aai_indices = indices

        #if input spectrum is none or empty, raise error.
        if use_dsp:
            if spectrum == None or spectrum == "":
                raise ValueError('Spectrum cannot be None or empty, got {}'.format(spectrum))

        #pandas series to store all output results
        aai_df = pd.Series(index=['Index','Category','R2', 'RMSE', 'MSE', 'RPD',
            'MAE', 'Explained Var'],dtype='object')

        #get AAI index encodings specified by indices input parameter
        encoded_seqs = self.get_aai_enoding(indices)

        #if use_dsp true then get protein spectra from AAI indices using ProDSP class
        #   else use the AAI indices encoding's themselves as the feature data (X)
        if use_dsp:
            proDSP = ProDSP(encoded_seqs, spectrum=spectrum, window=window, filter_=filter_)
            proDSP.encode_seqs()
            X = pd.DataFrame(proDSP.spectrum_encoding)
        else:
            X = pd.DataFrame(encoded_seqs)

        #get training and test dataset split
        X_train, X_test, Y_train, Y_test = self.model.train_test_split(X,
            self.get_activity(), test_size = self.test_split)

        #fit predictive model
        self.model.fit()

        #predict activity values for test data
        Y_pred = self.model.predict()

        #create instance of Evaluate class which will get all the evaluation metrics
        eval = Evaluate(Y_test, Y_pred)

        #get categories for all indices in self.aai_indices
        if isinstance(self.aai_indices, list):
            index_cat = ""
            for i in range(0,len(self.aai_indices)):
                index_cat += index_cat + ', '+ \
                    self.aaindex.get_category_from_record(self.aai_indices[i])
        else:
            index_cat = self.aaindex.get_category_from_record(self.aai_indices)

        #set results Series variables
        aai_df['Index'] = indices
        aai_df['Category'] = str(index_cat)
        aai_df['R2'] = eval.r2
        aai_df['RMSE'] = eval.rmse
        aai_df['MSE'] = eval.mse
        aai_df['RPD'] = eval.rpd
        aai_df['MAE'] = eval.mae
        aai_df['Explained Var'] = eval.explained_var

        #print out results from encoding
        self.output_results(aai_df)

        #plot regression plot for predictive model
        plot_reg(Y_test, Y_pred, eval.r2)

        #save results of encoding to output folder specified by OUTPUT_FOLDER
        save_results(aai_df, 'aai_encoding')

        #reset aai_indices instance variable
        # self.aai_indices = ""

        return aai_df

    def get_descriptor_encoding(self, descriptors=None):
        """
        Calculate inputted descriptor(s) requried for the encoding process.
        Get closest match to user inputted descriptor using difflib library.
        If a single descriptor is input then calculate it and return, if list of
        descriptors input then calculate descriptor values for each and concatenate.

        Parameters
        ----------
        descriptors : str/list (default = None)
            string or list of protein descriptors

        Returns
        -------
        encoded_desc : pd.DataFrame
            pandas dataFrame of calculated descriptor values according to user
            inputted descriptor.
        """
        #raise error if no descriptors specified in input
        if descriptors == None or descriptors == "":
            raise ValueError('No descriptors have been input.')

        #create instance of Descriptors class using data in instance variable
        descr = Descriptors(self.data[self.seq_col])

        #get closest valid available descriptors from input descriptor parameter,
        #   if a list of descriptors passed in as the input parameter then get
        #       all valid descriptors in list
        if isinstance(descriptors, list):
            for de in range(0,len(descriptors)):
                desc_matches = get_close_matches(descriptors[de],
                    descr.valid_descriptors(),cutoff=0.4)
                descriptors[de] = desc_matches[0]
                if descriptors[de] == []:
                    raise ValueError('No approximate descriptor found from one entered: {}'.format(de))
        else:
            desc_matches = get_close_matches(descriptors,descr.valid_descriptors(),cutoff=0.4)
            descriptors = desc_matches[0]
            if descriptors == []:
                raise ValueError('No approximate descriptor found from one entered: {}'.format(descriptors))

        #initialise temp lists and DF to store encoded descriptor values
        encoded_desc_temp = []
        encoded_desc_vals = []
        encoded_desc_temp = pd.DataFrame()

        #if single descriptor passed in, not a list, get descriptor values for sequences
        if not isinstance(descriptors, list):
            encoded_desc_temp = descr.get_descriptor_encoding(descriptors)

            #raise value error if descriptor is empty
            if (encoded_desc_temp.empty):
                raise ValueError('Descriptors {} cannot be empty or None'.format(descriptors))

            encoded_desc = encoded_desc_temp

            return encoded_desc

        #if list of descriptors passed as input, iterate and get each descriptors' values
        else:
            for d in range(0,len(descriptors)):
              encoded_desc_temp = descr.get_descriptor_encoding(descriptors[d])

              #raise value error if descriptor is empty
              if (encoded_desc_temp.empty):
                  raise ValueError('Descriptors {} cannot be empty or None'.format(descriptors[d]))
                    # self.descriptors[d]))

              encoded_desc_vals.append(encoded_desc_temp)
              encoded_desc_temp = pd.DataFrame()   #reset to empty dataframe

            #concatenate dataframes of descriptors
            encoded_desc = pd.concat(encoded_desc_vals,axis=1)

        return encoded_desc

    def encode_desc(self, descriptor=None):
        """
        Encode protein sequences using protein physiochemical and structural
        descriptors and build predictive model from the descriptor feature data.
        If multiple descriptors input then calculate each and concatenate them.
        The resulting model assets and its results will be exported to the
        directory pointed to by the global variable OUTPUT_DIR.

        Parameters
        ----------
        descriptor : str/list (default = None)
            string or list of protein descriptors

        Returns
        -------
        desc_df : pd.Series
            pandas series storing metrics and results of encoding.
        """
        #raise error if no descriptor specified in input
        if descriptor == None or descriptor == "":
            raise ValueError('No descriptors have been input, descriptors attribute is empty.')

        self.descriptors = descriptor

        #create instance of Descriptors class using data in instance variable
        descr = Descriptors(self.data[self.seq_col])

        #pandas series to store all output results
        desc_df = pd.Series(index=['Descriptor','Group','R2', 'RMSE', 'MSE',
            'RPD', 'MAE', 'Explained Var'],dtype='object')

        X = self.get_descriptor_encoding(descriptors = self.descriptors)

        #get training and test dataset split
        X_train, X_test, Y_train, Y_test = self.model.train_test_split(X,
            self.get_activity(), test_size = self.test_split)

        #fit predictive model
        self.model.fit()

        #predict activity values for test data
        Y_pred = self.model.predict()

        #create instance of Evaluate class which will get all the evaluation metrics
        eval = Evaluate(Y_test, Y_pred)

        desc_group = ""

        #get groups for all descriptors in self.desciptors
        if isinstance(self.descriptors, list):
            for i in range(0,len(self.descriptors)):
                desc_group += desc_group + ', '+ \
                    descr.descriptor_groups['_'+self.descriptors[i]]
        else:
            desc_cat = descr.descriptor_groups['_'+self.descriptors]

        #set results Series variables
        desc_df['Descriptor'] = descriptor
        desc_df['Group'] = desc_cat
        desc_df['R2'] = eval.r2
        desc_df['RMSE'] = eval.rmse
        desc_df['MSE'] = eval.mse
        desc_df['RPD'] = eval.rpd
        desc_df['MAE'] = eval.mae
        desc_df['Explained Var'] = eval.explained_var

        #print out results from encoding
        self.output_results(desc_df)

        #plot regression plot for predictive model
        plot_reg(Y_test, Y_pred, eval.r2)

        #save results of encoding to output folder
        save_results(desc_df, 'desc_encoding')

        #reset descriptors instance variable
        # self.descriptors = ""

        return desc_df

    def encode_aai_desc(self, indices=None,descriptors=None,spectrum='power',
        window="", filter_=None,use_dsp=True):
        """
        Encode using both AAI indices and the descriptors. The two outputs from
        the individual encoding strategies will be concatenated together and
        used in the building of a predictive model. The resulting model assets
        and its results will be exported to the directory pointed to by the global
        variable OUTPUT_DIR. If use_dsp is true then pass AAI Indices through a
        DSP transformation specified by parameters: spectrum, window & filter.

        Parameters
        ----------
        spectrum : str (default = "power")
            type of protein spectrum to use for DSP transformation.
        indices : str/list (default = None)
            string or list of indices from the AAI.
        descriptors : str/list (default = None)
            string or list of protein descriptors.
        window : str
            name of window function to apply to output of Fourier Transform.
        filter_ : str
            name of filter function to apply to ouput of Fourier Transform.
        use_dsp : bool (default = True)
            whether to transform AAI index encodings via DSP techniques and a Fourier
            transform. If false then no protein spectra will be generated and the
            normal encoded AAI indices will be used as the model feature data.

        Returns
        -------
        aai_desc_df : pd.Series
            pandas series storing metrics and results of encoding.
        """

        #validate AAI indices and Descriptors are present in the input parameters
        #   or instance variables, return error if either is None
        if (descriptors == None or descriptors == "") or \
            (indices == None or indices == ""):
                raise ValueError('AAI Indices and Descriptor input parameters must \
                    not be empty or None')

        self.aai_indices = indices           #set instance attributes
        self.descriptors = descriptors

        #create instance of Descriptors class using data in instance variable
        descr = Descriptors(self.data[self.seq_col])

        #create output results Series
        aai_desc_df = pd.Series(index=['Index','Category','Descriptor',
            'Group','R2', 'RMSE', 'MSE', 'RPD', 'MAE', 'Explained Var'],dtype='object')

        #get AAI index encoding features
        aai_encoding = self.get_aai_enoding(indices)
        aai_encoding = pd.DataFrame(aai_encoding)

        #if AAI indices encoding is empty, raise error
        if (aai_encoding.empty):
            raise ValueError('AAI Indices {} cannot be empty or None'.format(indices))

        #get descriptor encoding features
        # descriptor_encoding = self.desc_encoding(descriptors)
        descriptor_encoding = self.get_descriptor_encoding(descriptors)

        #if descriptors encoding is empty, raise error
        if (descriptor_encoding.empty):
            raise ValueError('Descriptors {} cannot be empty or None'.format(descriptors))

        #concatenate AAI index and Descriptor features
        aai_desc = [aai_encoding, descriptor_encoding]
        X = pd.concat(aai_desc, axis = 1)

        #get training and test dataset split
        X_train, X_test, Y_train, Y_test = self.model.train_test_split(X,
            self.get_activity(), test_size = self.test_split)

        #fit predictive model
        self.model.fit()

        #predict activity values for test data
        Y_pred = self.model.predict()

        #create instance of Evaluate class which will get all the evaluation metrics
        eval = Evaluate(Y_test, Y_pred)

        index_cat = ""

        #get categories for all indices in self.aai_indices
        if isinstance(self.aai_indices, list):
            for i in range(0,len(self.aai_indices)):
                index_cat += index_cat + ', '+ \
                    self.aaindex.get_category_from_record(self.aai_indices[i])
        else:
            index_cat = self.aaindex.get_category_from_record(self.aai_indices)

        desc_group = ""

        #get groups for all descriptors in self.desciptors
        if isinstance(self.descriptors, list):
            for i in range(0,len(self.descriptors)):
                desc_group += desc_group + ', '+ \
                    descr.descriptor_groups[self.descriptors[i]]
        else:
            desc_cat = descr.descriptor_groups[self.descriptors]

        #set output dataframe columns
        aai_desc_df['Index'] = str(self.aai_indices)
        aai_desc_df['Category'] = str(index_cat)
        aai_desc_df['Descriptor'] = str(self.descriptors)
        aai_desc_df['Group'] = str(desc_cat)
        aai_desc_df['R2'] = eval.r2
        aai_desc_df['RMSE'] = eval.rmse
        aai_desc_df['MSE'] = eval.mse
        aai_desc_df['RPD'] = eval.rpd
        aai_desc_df['MAE'] = eval.mae
        aai_desc_df['Explained Var'] = eval.explained_var

        #print out results from encoding
        self.output_results(aai_desc_df)

        #plot regression plot for predictive model
        plot_reg(Y_test, Y_pred, eval.r2)

        #save results of encoding to output folder
        save_results(aai_desc_df, 'aai_desc_encoding')

        return aai_desc_df

    def output_results(self, results):
        """
        Print out the predictive model parameters/attributes and its results.

        Parameters
        ----------
        results : dict/pd.Series
            dictionary or Series of metrics and their associated values.
        """
        print('\n#############################################################')
        print('######################## Results ############################\n')
        print('#############################################################\n')
        print('####################### Parameters ##########################\n')
        print('# Dataset -> {}\n# Dataset Size -> {}\n# Sequence Length -> {} \
            \n# Activity -> {}\n# AAI Indices -> {}\n# Descriptors -> {}\n# Algorithm -> {}\
                \n# Model Parameters -> {}\n# Test Split -> {}\n'.format(
                self.dataset,self.num_seqs, self.seq_len, self.activity, self.aai_indices,
                self.descriptors, repr(self.model), self.model.model.get_params(),
                self.test_split
                ))
        print('######################## Metrics ############################\n')
        print('# R2: {}'.format(results['R2']))
        print('# RMSE: {} '.format(results['RMSE']))
        print('# MSE: {} '.format(results['MSE']))
        print('# MAE: {}'.format(results['MAE']))
        print('# RPD {}'.format(results['RPD']))
        print('# Variance {}\n'.format(results['Explained Var']))
        print('#############################################################\n')

    def get_seqs(self):
        """
        Return all protein sequences within the dataset.

        Returns
        -------
        self.data[self.seq_col] : pd.Series
            pandas series of all protein sequences in the dataset.
        """
        return self.data[self.seq_col]

    def get_activity(self):
        """
        Return all activity/fitness values for the sequences within the dataset.

        Returns
        -------
        self.data[self.activity] : np.ndarray
            array of all activity/fitness values for the protein sequences
        """
        return self.data[self.activity].values.reshape((-1,1))

######################          Getters & Setters          ######################

    @property
    def dataset(self):
        return self._dataset

    @dataset.setter
    def dataset(self, val):
        self._dataset = val

    @property
    def seq_col(self):
        return self._seq_col

    @seq_col.setter
    def seq_col(self, val):
        self._seq_col = val

    @property
    def activity(self):
        return self._activity

    @activity.setter
    def activity(self, val):
        self._activity = val

    @property
    def aai_indices(self):
        return self._aai_indices

    @aai_indices.setter
    def aai_indices(self, val):
        self._aai_indices = val

    @property
    def descriptors(self):
        return self._descriptors

    @descriptors.setter
    def descriptors(self, val):
        self._descriptors = val

    @property
    def algorithm(self):
        return self._algorithm

    @algorithm.setter
    def algorithm(self, val):
        self._algorithm = val

    @property
    def parameters(self):
        return self._parameters

    @parameters.setter
    def parameters(self, val):
        self._parameters = val

    @property
    def test_split(self):
        return self._test_split

    @test_split.setter
    def test_split(self, val):
        self._test_split = val

    @property
    def num_seqs(self):
        return self._num_seqs

    @num_seqs.setter
    def num_seqs(self, val):
        self._num_seqs = val

    @property
    def seq_len(self):
        return self._seq_len

    @seq_len.setter
    def seq_len(self, val):
        self._seq_len = val

################################################################################

    def __str__(self):
        return "Instance of PySAR class, using parameters: {}".format(self.__dict__)

    def __repr__(self):
        return "<PySAR: {} >".format(self)



# -*- coding: utf-8 -*-
#  Copyright (c) 2016-2017, Zhijiang Yao, Jie Dong and Dongsheng Cao
#  All rights reserved.
#  This file is part of the PyBioMed.
#  The contents are covered by the terms of the BSD license
#  which is included in the file license.txt, found at the root
#  of the PyBioMed source tree.
"""
###############################################################################

The module is used for computing the composition of amino acids, dipetide and

3-mers (tri-peptide) for a given protein sequence. You can get 8420 descriptors

for a given protein sequence. You can freely use and distribute it. If you hava

any problem, you could contact with us timely!

References:

[1]: Reczko, M. and Bohr, H. (1994) The DEF data base of sequence based protein

fold class predictions. Nucleic Acids Res, 22, 3616-3619.

[2]: Hua, S. and Sun, Z. (2001) Support vector machine approach for protein

subcellular localization prediction. Bioinformatics, 17, 721-728.


[3]:Grassmann, J., Reczko, M., Suhai, S. and Edler, L. (1999) Protein fold class

prediction: new methods of statistical classification. Proc Int Conf Intell Syst Mol

Biol, 106-112.

Authors: Zhijiang Yao and Dongsheng Cao.

Date: 2016.06.04

Email: gadsby@163.com

###############################################################################
"""

# Core Library modules
import re

AALetter = [
    "A",
    "R",
    "N",
    "D",
    "C",
    "E",
    "Q",
    "G",
    "H",
    "I",
    "L",
    "K",
    "M",
    "F",
    "P",
    "S",
    "T",
    "W",
    "Y",
    "V",
]


#############################################################################################
def CalculateAAComposition(ProteinSequence):
    """
    ########################################################################
    Calculate the composition of Amino acids

    for a given protein sequence.

    Usage:

    result=CalculateAAComposition(protein)

    Input: protein is a pure protein sequence.

    Output: result is a dict form containing the composition of

    20 amino acids.
    ########################################################################
    """
    LengthSequence = len(ProteinSequence)
    Result = {}
    for i in AALetter:
        Result[i] = round(float(ProteinSequence.count(i)) / LengthSequence * 100, 3)
    return Result


#############################################################################################
def CalculateDipeptideComposition(ProteinSequence):
    """
    ########################################################################
    Calculate the composition of dipeptidefor a given protein sequence.

    Usage:

    result=CalculateDipeptideComposition(protein)

    Input: protein is a pure protein sequence.

    Output: result is a dict form containing the composition of

    400 dipeptides.
    ########################################################################
    """

    LengthSequence = len(ProteinSequence)
    Result = {}
    for i in AALetter:
        for j in AALetter:
            Dipeptide = i + j
            Result[Dipeptide] = round(
                float(ProteinSequence.count(Dipeptide)) / (LengthSequence - 1) * 100, 2
            )
    return Result


#############################################################################################


def Getkmers():
    """
    ########################################################################
    Get the amino acid list of 3-mers.

    Usage:

    result=Getkmers()

    Output: result is a list form containing 8000 tri-peptides.

    ########################################################################
    """
    kmers = list()
    for i in AALetter:
        for j in AALetter:
            for k in AALetter:
                kmers.append(i + j + k)
    return kmers


#############################################################################################
def GetSpectrumDict(proteinsequence):
    """
    ########################################################################
    Calcualte the spectrum descriptors of 3-mers for a given protein.

    Usage:

    result=GetSpectrumDict(protein)

    Input: protein is a pure protein sequence.

    Output: result is a dict form containing the composition values of 8000

    3-mers.
    ########################################################################
    """
    result = {}
    kmers = Getkmers()
    for i in kmers:
        result[i] = len(re.findall(i, proteinsequence))
    return result


#############################################################################################
def CalculateAADipeptideComposition(ProteinSequence):
    """
    ########################################################################
    Calculate the composition of AADs, dipeptide and 3-mers for a

    given protein sequence.

    Usage:

    result=CalculateAADipeptideComposition(protein)

    Input: protein is a pure protein sequence.

    Output: result is a dict form containing all composition values of

    AADs, dipeptide and 3-mers (8420).
    ########################################################################
    """

    result = {}
    result.update(CalculateAAComposition(ProteinSequence))
    result.update(CalculateDipeptideComposition(ProteinSequence))
    result.update(GetSpectrumDict(ProteinSequence))

    return result


#############################################################################################
if __name__ == "__main__":
    protein = "ADGCGVGEGTGQGPMCNCMCMKWVYADEDAADLESDSFADEDASLESDSFPWSNQRVFCSFADEDAS"

    AAC = CalculateAAComposition(protein)
    print(AAC)
    DIP = CalculateDipeptideComposition(protein)
    print(DIP)
    spectrum = GetSpectrumDict(protein)
    print(spectrum)
    res = CalculateAADipeptideComposition(protein)
    print(len(res))



# -*- coding: utf-8 -*-
#  Copyright (c) 2016-2017, Zhijiang Yao, Jie Dong and Dongsheng Cao
#  All rights reserved.
#  This file is part of the PyBioMed.
#  The contents are covered by the terms of the BSD license
#  which is included in the file license.txt, found at the root
#  of the PyBioMed source tree.
"""
##########################################################################################

This module is used for computing the Autocorrelation descriptors based different

 properties of AADs.You can also input your properties of AADs, then it can help you

to compute Autocorrelation descriptors based on the property of AADs. Currently, You

can get 720 descriptors for a given protein sequence based on our provided physicochemical

properties of AADs. You can freely use and distribute it. If you hava  any problem,

you could contact with us timely!

References:

[1]: http://www.genome.ad.jp/dbget/aaindex.html

[2]:Feng, Z.P. and Zhang, C.T. (2000) Prediction of membrane protein types based on

the hydrophobic index of amino acids. J Protein Chem, 19, 269-275.

[3]:Horne, D.S. (1988) Prediction of protein helix content from an autocorrelation

analysis of sequence hydrophobicities. Biopolymers, 27, 451-477.

[4]:Sokal, R.R. and Thomson, B.A. (2006) Population structure inferred by local

spatial autocorrelation: an Usage from an Amerindian tribal population. Am J

Phys Anthropol, 129, 121-131.

Authors: Zhijiang Yao and Dongsheng Cao.

Date: 2016.06.04

Email: gadsby@163.com

##########################################################################################
"""

# Core Library modules
import math
import string

AALetter = [
    "A",
    "R",
    "N",
    "D",
    "C",
    "Q",
    "E",
    "G",
    "H",
    "I",
    "L",
    "K",
    "M",
    "F",
    "P",
    "S",
    "T",
    "W",
    "Y",
    "V",
]

_Hydrophobicity = {
    "A": 0.02,
    "R": -0.42,
    "N": -0.77,
    "D": -1.04,
    "C": 0.77,
    "Q": -1.10,
    "E": -1.14,
    "G": -0.80,
    "H": 0.26,
    "I": 1.81,
    "L": 1.14,
    "K": -0.41,
    "M": 1.00,
    "F": 1.35,
    "P": -0.09,
    "S": -0.97,
    "T": -0.77,
    "W": 1.71,
    "Y": 1.11,
    "V": 1.13,
}

_AvFlexibility = {
    "A": 0.357,
    "R": 0.529,
    "N": 0.463,
    "D": 0.511,
    "C": 0.346,
    "Q": 0.493,
    "E": 0.497,
    "G": 0.544,
    "H": 0.323,
    "I": 0.462,
    "L": 0.365,
    "K": 0.466,
    "M": 0.295,
    "F": 0.314,
    "P": 0.509,
    "S": 0.507,
    "T": 0.444,
    "W": 0.305,
    "Y": 0.420,
    "V": 0.386,
}

_Polarizability = {
    "A": 0.046,
    "R": 0.291,
    "N": 0.134,
    "D": 0.105,
    "C": 0.128,
    "Q": 0.180,
    "E": 0.151,
    "G": 0.000,
    "H": 0.230,
    "I": 0.186,
    "L": 0.186,
    "K": 0.219,
    "M": 0.221,
    "F": 0.290,
    "P": 0.131,
    "S": 0.062,
    "T": 0.108,
    "W": 0.409,
    "Y": 0.298,
    "V": 0.140,
}

_FreeEnergy = {
    "A": -0.368,
    "R": -1.03,
    "N": 0.0,
    "D": 2.06,
    "C": 4.53,
    "Q": 0.731,
    "E": 1.77,
    "G": -0.525,
    "H": 0.0,
    "I": 0.791,
    "L": 1.07,
    "K": 0.0,
    "M": 0.656,
    "F": 1.06,
    "P": -2.24,
    "S": -0.524,
    "T": 0.0,
    "W": 1.60,
    "Y": 4.91,
    "V": 0.401,
}

_ResidueASA = {
    "A": 115.0,
    "R": 225.0,
    "N": 160.0,
    "D": 150.0,
    "C": 135.0,
    "Q": 180.0,
    "E": 190.0,
    "G": 75.0,
    "H": 195.0,
    "I": 175.0,
    "L": 170.0,
    "K": 200.0,
    "M": 185.0,
    "F": 210.0,
    "P": 145.0,
    "S": 115.0,
    "T": 140.0,
    "W": 255.0,
    "Y": 230.0,
    "V": 155.0,
}

_ResidueVol = {
    "A": 52.6,
    "R": 109.1,
    "N": 75.7,
    "D": 68.4,
    "C": 68.3,
    "Q": 89.7,
    "E": 84.7,
    "G": 36.3,
    "H": 91.9,
    "I": 102.0,
    "L": 102.0,
    "K": 105.1,
    "M": 97.7,
    "F": 113.9,
    "P": 73.6,
    "S": 54.9,
    "T": 71.2,
    "W": 135.4,
    "Y": 116.2,
    "V": 85.1,
}

_Steric = {
    "A": 0.52,
    "R": 0.68,
    "N": 0.76,
    "D": 0.76,
    "C": 0.62,
    "Q": 0.68,
    "E": 0.68,
    "G": 0.00,
    "H": 0.70,
    "I": 1.02,
    "L": 0.98,
    "K": 0.68,
    "M": 0.78,
    "F": 0.70,
    "P": 0.36,
    "S": 0.53,
    "T": 0.50,
    "W": 0.70,
    "Y": 0.70,
    "V": 0.76,
}

_Mutability = {
    "A": 100.0,
    "R": 65.0,
    "N": 134.0,
    "D": 106.0,
    "C": 20.0,
    "Q": 93.0,
    "E": 102.0,
    "G": 49.0,
    "H": 66.0,
    "I": 96.0,
    "L": 40.0,
    "K": -56.0,
    "M": 94.0,
    "F": 41.0,
    "P": 56.0,
    "S": 120.0,
    "T": 97.0,
    "W": 18.0,
    "Y": 41.0,
    "V": 74.0,
}


###You can continuely add other properties of AADs to compute the descriptors of protein sequence.


_AAProperty = (
    _Hydrophobicity,
    _AvFlexibility,
    _Polarizability,
    _FreeEnergy,
    _ResidueASA,
    _ResidueVol,
    _Steric,
    _Mutability,
)

_AAPropertyName = (
    "_Hydrophobicity",
    "_AvFlexibility",
    "_Polarizability",
    "_FreeEnergy",
    "_ResidueASA",
    "_ResidueVol",
    "_Steric",
    "_Mutability",
)


##################################################################################################
def _mean(listvalue):
    """
    The mean value of the list data.
    """
    return sum(listvalue) / len(listvalue)


##################################################################################################
def _std(listvalue, ddof=1):
    """
    The standard deviation of the list data.
    """
    mean = _mean(listvalue)
    temp = [math.pow(i - mean, 2) for i in listvalue]
    res = math.sqrt(sum(temp) / (len(listvalue) - ddof))
    return res


##################################################################################################


def NormalizeEachAAP(AAP):
    """
    ####################################################################################
    All of the amino acid indices are centralized and

    standardized before the calculation.

    Usage:

    result=NormalizeEachAAP(AAP)

    Input: AAP is a dict form containing the properties of 20 amino acids.

    Output: result is the a dict form containing the normalized properties

    of 20 amino acids.
    ####################################################################################
    """
    if len(AAP.values()) != 20:
        print("You can not input the correct number of properities of Amino acids!")
    else:
        Result = {}
        for i, j in AAP.items():
            Result[i] = (j - _mean(AAP.values())) / _std(AAP.values(), ddof=0)

    return Result


def CalculateEachNormalizedMoreauBrotoAuto(ProteinSequence, AAP, AAPName):
    """
    ####################################################################################
    you can use the function to compute MoreauBrotoAuto

    descriptors for different properties based on AADs.

    Usage:

    result=CalculateEachNormalizedMoreauBrotoAuto(protein,AAP,AAPName)

    Input: protein is a pure protein sequence.

    AAP is a dict form containing the properties of 20 amino acids (e.g., _AvFlexibility).

    AAPName is a string used for indicating the property (e.g., '_AvFlexibility').

    Output: result is a dict form containing 30 Normalized Moreau-Broto autocorrelation

    descriptors based on the given property.
    ####################################################################################
    """

    AAPdic = NormalizeEachAAP(AAP)

    Result = {}
    for i in range(1, 31):
        temp = 0
        for j in range(len(ProteinSequence) - i):
            temp = temp + AAPdic[ProteinSequence[j]] * AAPdic[ProteinSequence[j + 1]]
        if len(ProteinSequence) - i == 0:
            Result["MoreauBrotoAuto" + AAPName + str(i)] = round(
                temp / (len(ProteinSequence)), 3
            )
        else:
            Result["MoreauBrotoAuto" + AAPName + str(i)] = round(
                temp / (len(ProteinSequence) - i), 3
            )

    return Result


def CalculateEachMoranAuto(ProteinSequence, AAP, AAPName):
    """
    ####################################################################################
    you can use the function to compute MoranAuto

    descriptors for different properties based on AADs.

    Usage:

    result=CalculateEachMoranAuto(protein,AAP,AAPName)

    Input: protein is a pure protein sequence.

    AAP is a dict form containing the properties of 20 amino acids (e.g., _AvFlexibility).

    AAPName is a string used for indicating the property (e.g., '_AvFlexibility').

    Output: result is a dict form containing 30 Moran autocorrelation

    descriptors based on the given property.
    ####################################################################################
    """

    AAPdic = NormalizeEachAAP(AAP)

    cds = 0
    for i in AALetter:
        cds = cds + (ProteinSequence.count(i)) * (AAPdic[i])
    Pmean = cds / len(ProteinSequence)

    cc = []
    for i in ProteinSequence:
        cc.append(AAPdic[i])

    K = (_std(cc, ddof=0)) ** 2

    Result = {}
    for i in range(1, 31):
        temp = 0
        for j in range(len(ProteinSequence) - i):
            temp = temp + (AAPdic[ProteinSequence[j]] - Pmean) * (
                AAPdic[ProteinSequence[j + i]] - Pmean
            )
        if len(ProteinSequence) - i == 0:
            Result["MoranAuto" + AAPName + str(i)] = round(
                temp / (len(ProteinSequence)) / K, 3
            )
        else:
            Result["MoranAuto" + AAPName + str(i)] = round(
                temp / (len(ProteinSequence) - i) / K, 3
            )

    return Result


def CalculateEachGearyAuto(ProteinSequence, AAP, AAPName):
    """
    ####################################################################################
    you can use the function to compute GearyAuto

    descriptors for different properties based on AADs.

    Usage:

    result=CalculateEachGearyAuto(protein,AAP,AAPName)

    Input: protein is a pure protein sequence.

    AAP is a dict form containing the properties of 20 amino acids (e.g., _AvFlexibility).

    AAPName is a string used for indicating the property (e.g., '_AvFlexibility').

    Output: result is a dict form containing 30 Geary autocorrelation

    descriptors based on the given property.
    ####################################################################################
    """

    AAPdic = NormalizeEachAAP(AAP)

    cc = []
    for i in ProteinSequence:
        cc.append(AAPdic[i])

    K = ((_std(cc)) ** 2) * len(ProteinSequence) / (len(ProteinSequence) - 1)
    Result = {}
    for i in range(1, 31):
        temp = 0
        for j in range(len(ProteinSequence) - i):
            temp = (
                temp
                + (AAPdic[ProteinSequence[j]] - AAPdic[ProteinSequence[j + i]]) ** 2
            )
        if len(ProteinSequence) - i == 0:
            Result["GearyAuto" + AAPName + str(i)] = round(
                temp / (2 * (len(ProteinSequence))) / K, 3
            )
        else:
            Result["GearyAuto" + AAPName + str(i)] = round(
                temp / (2 * (len(ProteinSequence) - i)) / K, 3
            )
    return Result


##################################################################################################


def CalculateNormalizedMoreauBrotoAuto(ProteinSequence, AAProperty, AAPropertyName):
    """
    ####################################################################################
    A method used for computing MoreauBrotoAuto for all properties.

    Usage:

    result=CalculateNormalizedMoreauBrotoAuto(protein,AAP,AAPName)

    Input: protein is a pure protein sequence.

    AAProperty is a list or tuple form containing the properties of 20 amino acids (e.g., _AAProperty).

    AAPName is a list or tuple form used for indicating the property (e.g., '_AAPropertyName').

    Output: result is a dict form containing 30*p Normalized Moreau-Broto autocorrelation

    descriptors based on the given properties.
    ####################################################################################

    """
    Result = {}
    for i in range(len(AAProperty)):
        Result[AAPropertyName[i]] = CalculateEachNormalizedMoreauBrotoAuto(
            ProteinSequence, AAProperty[i], AAPropertyName[i]
        )

    return Result


def CalculateMoranAuto(ProteinSequence, AAProperty, AAPropertyName):
    """
    ####################################################################################
    A method used for computing MoranAuto for all properties

    Usage:

    result=CalculateMoranAuto(protein,AAP,AAPName)

    Input: protein is a pure protein sequence.

    AAProperty is a list or tuple form containing the properties of 20 amino acids (e.g., _AAProperty).

    AAPName is a list or tuple form used for indicating the property (e.g., '_AAPropertyName').

    Output: result is a dict form containing 30*p Moran autocorrelation

    descriptors based on the given properties.
    ####################################################################################
    """
    Result = {}
    for i in range(len(AAProperty)):
        Result[AAPropertyName[i]] = CalculateEachMoranAuto(
            ProteinSequence, AAProperty[i], AAPropertyName[i]
        )

    return Result


def CalculateGearyAuto(ProteinSequence, AAProperty, AAPropertyName):
    """
    ####################################################################################
    A method used for computing GearyAuto for all properties

    Usage:

    result=CalculateGearyAuto(protein,AAP,AAPName)

    Input: protein is a pure protein sequence.

    AAProperty is a list or tuple form containing the properties of 20 amino acids (e.g., _AAProperty).

    AAPName is a list or tuple form used for indicating the property (e.g., '_AAPropertyName').

    Output: result is a dict form containing 30*p Geary autocorrelation

    descriptors based on the given properties.
    ####################################################################################
    """
    Result = {}
    for i in range(len(AAProperty)):
        Result[AAPropertyName[i]] = CalculateEachGearyAuto(
            ProteinSequence, AAProperty[i], AAPropertyName[i]
        )

    return Result


########################NormalizedMoreauBorto##################################
def CalculateNormalizedMoreauBrotoAutoHydrophobicity(ProteinSequence):
    """
    ####################################################################################
    Calculte the NormalizedMoreauBorto Autocorrelation descriptors based on

    hydrophobicity.

    Usage:

    result=CalculateNormalizedMoreauBrotoAutoHydrophobicity(protein)

    Input: protein is a pure protein sequence.

    Output: result is a dict form containing 30 Normalized Moreau-Broto Autocorrelation

    descriptors based on Hydrophobicity.
    ####################################################################################
    """

    result = CalculateEachNormalizedMoreauBrotoAuto(
        ProteinSequence, _Hydrophobicity, "_Hydrophobicity"
    )
    return result


def CalculateNormalizedMoreauBrotoAutoAvFlexibility(ProteinSequence):
    """
    ####################################################################################
    Calculte the NormalizedMoreauBorto Autocorrelation descriptors based on

    AvFlexibility.

    Usage:

    result=CalculateNormalizedMoreauBrotoAutoAvFlexibility(protein)

    Input: protein is a pure protein sequence.

    Output: result is a dict form containing 30 Normalized Moreau-Broto Autocorrelation

    descriptors based on AvFlexibility.
    ####################################################################################
    """

    result = CalculateEachNormalizedMoreauBrotoAuto(
        ProteinSequence, _AvFlexibility, "_AvFlexibility"
    )
    return result


def CalculateNormalizedMoreauBrotoAutoPolarizability(ProteinSequence):
    """
    ####################################################################################
    Calculte the NormalizedMoreauBorto Autocorrelation descriptors based on

    Polarizability.

    Usage:

    result=CalculateNormalizedMoreauBrotoAutoPolarizability(protein)

    Input: protein is a pure protein sequence.

    Output: result is a dict form containing 30 Normalized Moreau-Broto Autocorrelation

    descriptors based on Polarizability.
    ####################################################################################
    """

    result = CalculateEachNormalizedMoreauBrotoAuto(
        ProteinSequence, _Polarizability, "_Polarizability"
    )
    return result


def CalculateNormalizedMoreauBrotoAutoFreeEnergy(ProteinSequence):
    """
    ####################################################################################
    Calculte the NormalizedMoreauBorto Autocorrelation descriptors based on

    FreeEnergy.

    Usage:

    result=CalculateNormalizedMoreauBrotoAutoFreeEnergy(protein)

    Input: protein is a pure protein sequence.

    Output: result is a dict form containing 30 Normalized Moreau-Broto Autocorrelation

    descriptors based on FreeEnergy.
    ####################################################################################
    """

    result = CalculateEachNormalizedMoreauBrotoAuto(
        ProteinSequence, _FreeEnergy, "_FreeEnergy"
    )
    return result


def CalculateNormalizedMoreauBrotoAutoResidueASA(ProteinSequence):
    """
    ####################################################################################
    Calculte the NormalizedMoreauBorto Autocorrelation descriptors based on

    ResidueASA.

    Usage:

    result=CalculateNormalizedMoreauBrotoAutoResidueASA(protein)

    Input: protein is a pure protein sequence.

    Output: result is a dict form containing 30 Normalized Moreau-Broto Autocorrelation

    descriptors based on ResidueASA.
    ####################################################################################
    """

    result = CalculateEachNormalizedMoreauBrotoAuto(
        ProteinSequence, _ResidueASA, "_ResidueASA"
    )
    return result


def CalculateNormalizedMoreauBrotoAutoResidueVol(ProteinSequence):
    """
    ####################################################################################
    Calculte the NormalizedMoreauBorto Autocorrelation descriptors based on

    ResidueVol.

    Usage:

    result=CalculateNormalizedMoreauBrotoAutoResidueVol(protein)

    Input: protein is a pure protein sequence.

    Output: result is a dict form containing 30 Normalized Moreau-Broto Autocorrelation

    descriptors based on ResidueVol.
    ####################################################################################
    """

    result = CalculateEachNormalizedMoreauBrotoAuto(
        ProteinSequence, _ResidueVol, "_ResidueVol"
    )
    return result


def CalculateNormalizedMoreauBrotoAutoSteric(ProteinSequence):
    """
    ####################################################################################
    Calculte the NormalizedMoreauBorto Autocorrelation descriptors based on Steric.

    Usage:

    result=CalculateNormalizedMoreauBrotoAutoSteric(protein)

    Input: protein is a pure protein sequence.

    Output: result is a dict form containing 30 Normalized Moreau-Broto Autocorrelation

    descriptors based on Steric.
    ####################################################################################
    """

    result = CalculateEachNormalizedMoreauBrotoAuto(ProteinSequence, _Steric, "_Steric")
    return result


def CalculateNormalizedMoreauBrotoAutoMutability(ProteinSequence):
    """
    ####################################################################################
    Calculte the NormalizedMoreauBorto Autocorrelation descriptors based on Mutability.

    Usage:

    result=CalculateNormalizedMoreauBrotoAutoMutability(protein)

    Input: protein is a pure protein sequence.

    Output: result is a dict form containing 30 Normalized Moreau-Broto Autocorrelation

    descriptors based on Mutability.
    ####################################################################################
    """

    result = CalculateEachNormalizedMoreauBrotoAuto(
        ProteinSequence, _Mutability, "_Mutability"
    )
    return result


############################################################################

##############################MoranAuto######################################
def CalculateMoranAutoHydrophobicity(ProteinSequence):
    """
    ####################################################################################
    Calculte the MoranAuto Autocorrelation descriptors based on hydrophobicity.

    Usage:

    result=CalculateMoranAutoHydrophobicity(protein)

    Input: protein is a pure protein sequence.

    Output: result is a dict form containing 30 Moran Autocorrelation

    descriptors based on hydrophobicity.
    ####################################################################################
    """

    result = CalculateEachMoranAuto(ProteinSequence, _Hydrophobicity, "_Hydrophobicity")
    return result


def CalculateMoranAutoAvFlexibility(ProteinSequence):
    """
    ####################################################################################
    Calculte the MoranAuto Autocorrelation descriptors based on

    AvFlexibility.

    Usage:

    result=CalculateMoranAutoAvFlexibility(protein)

    Input: protein is a pure protein sequence.

    Output: result is a dict form containing 30 Moran Autocorrelation

    descriptors based on AvFlexibility.
    ####################################################################################
    """

    result = CalculateEachMoranAuto(ProteinSequence, _AvFlexibility, "_AvFlexibility")
    return result


def CalculateMoranAutoPolarizability(ProteinSequence):
    """
    ####################################################################################
    Calculte the MoranAuto Autocorrelation descriptors based on

    Polarizability.

    Usage:

    result=CalculateMoranAutoPolarizability(protein)

    Input: protein is a pure protein sequence.

    Output: result is a dict form containing 30 Moran Autocorrelation

    descriptors based on Polarizability.
    ####################################################################################
    """

    result = CalculateEachMoranAuto(ProteinSequence, _Polarizability, "_Polarizability")
    return result


def CalculateMoranAutoFreeEnergy(ProteinSequence):
    """
    ####################################################################################
    Calculte the MoranAuto Autocorrelation descriptors based on

    FreeEnergy.

    Usage:

    result=CalculateMoranAutoFreeEnergy(protein)

    Input: protein is a pure protein sequence.

    Output: result is a dict form containing 30 Moran Autocorrelation

    descriptors based on FreeEnergy.
    ####################################################################################
    """

    result = CalculateEachMoranAuto(ProteinSequence, _FreeEnergy, "_FreeEnergy")
    return result


def CalculateMoranAutoResidueASA(ProteinSequence):
    """
    ####################################################################################
    Calculte the MoranAuto Autocorrelation descriptors based on

    ResidueASA.

    Usage:

    result=CalculateMoranAutoResidueASA(protein)

    Input: protein is a pure protein sequence.

    Output: result is a dict form containing 30 Moran Autocorrelation

    descriptors based on ResidueASA.
    ####################################################################################
    """

    result = CalculateEachMoranAuto(ProteinSequence, _ResidueASA, "_ResidueASA")
    return result


def CalculateMoranAutoResidueVol(ProteinSequence):
    """
    ####################################################################################
    Calculte the MoranAuto Autocorrelation descriptors based on

    ResidueVol.

    Usage:

    result=CalculateMoranAutoResidueVol(protein)

    Input: protein is a pure protein sequence.

    Output: result is a dict form containing 30 Moran Autocorrelation

    descriptors based on ResidueVol.
    ####################################################################################
    """

    result = CalculateEachMoranAuto(ProteinSequence, _ResidueVol, "_ResidueVol")
    return result


def CalculateMoranAutoSteric(ProteinSequence):
    """
    ####################################################################################
    Calculte the MoranAuto Autocorrelation descriptors based on

    AutoSteric.

    Usage:

    result=CalculateMoranAutoSteric(protein)

    Input: protein is a pure protein sequence.

    Output: result is a dict form containing 30 Moran Autocorrelation

    descriptors based on AutoSteric.
    ####################################################################################
    """

    result = CalculateEachMoranAuto(ProteinSequence, _Steric, "_Steric")
    return result


def CalculateMoranAutoMutability(ProteinSequence):
    """
    ####################################################################################
    Calculte the MoranAuto Autocorrelation descriptors based on

    Mutability.

    Usage:

    result=CalculateMoranAutoMutability(protein)

    Input: protein is a pure protein sequence.

    Output: result is a dict form containing 30 Moran Autocorrelation

    descriptors based on Mutability.
    ####################################################################################
    """

    result = CalculateEachMoranAuto(ProteinSequence, _Mutability, "_Mutability")
    return result


############################################################################


################################GearyAuto#####################################
def CalculateGearyAutoHydrophobicity(ProteinSequence):
    """
    ####################################################################################
    Calculte the GearyAuto Autocorrelation descriptors based on

    hydrophobicity.

    Usage:

    result=CalculateGearyAutoHydrophobicity(protein)

    Input: protein is a pure protein sequence.

    Output: result is a dict form containing 30 Geary Autocorrelation

    descriptors based on hydrophobicity.
    ####################################################################################
    """

    result = CalculateEachGearyAuto(ProteinSequence, _Hydrophobicity, "_Hydrophobicity")
    return result


def CalculateGearyAutoAvFlexibility(ProteinSequence):
    """
    ####################################################################################
    Calculte the GearyAuto Autocorrelation descriptors based on

    AvFlexibility.

    Usage:
    result=CalculateGearyAutoAvFlexibility(protein)

    Input: protein is a pure protein sequence.

    Output: result is a dict form containing 30 Geary Autocorrelation

    descriptors based on AvFlexibility.
    ####################################################################################
    """

    result = CalculateEachGearyAuto(ProteinSequence, _AvFlexibility, "_AvFlexibility")
    return result


def CalculateGearyAutoPolarizability(ProteinSequence):
    """
    ####################################################################################
    Calculte the GearyAuto Autocorrelation descriptors based on

    Polarizability.

    Usage:

    result=CalculateGearyAutoPolarizability(protein)

    Input: protein is a pure protein sequence.

    Output: result is a dict form containing 30 Geary Autocorrelation

    descriptors based on Polarizability.
    ####################################################################################
    """

    result = CalculateEachGearyAuto(ProteinSequence, _Polarizability, "_Polarizability")
    return result


def CalculateGearyAutoFreeEnergy(ProteinSequence):
    """
    ####################################################################################
    Calculte the GearyAuto Autocorrelation descriptors based on

    FreeEnergy.

    Usage:

    result=CalculateGearyAutoFreeEnergy(protein)

    Input: protein is a pure protein sequence.

    Output: result is a dict form containing 30 Geary Autocorrelation

    descriptors based on FreeEnergy.
    ####################################################################################
    """

    result = CalculateEachGearyAuto(ProteinSequence, _FreeEnergy, "_FreeEnergy")
    return result


def CalculateGearyAutoResidueASA(ProteinSequence):
    """
    ####################################################################################
    Calculte the GearyAuto Autocorrelation descriptors based on

    ResidueASA.

    Usage:

    result=CalculateGearyAutoResidueASA(protein)

    Input: protein is a pure protein sequence.

    Output: result is a dict form containing 30 Geary Autocorrelation

    descriptors based on ResidueASA.
    ####################################################################################
    """

    result = CalculateEachGearyAuto(ProteinSequence, _ResidueASA, "_ResidueASA")
    return result


def CalculateGearyAutoResidueVol(ProteinSequence):
    """
    ####################################################################################
    Calculte the GearyAuto Autocorrelation descriptors based on

    ResidueVol.

    Usage:

    result=CalculateGearyAutoResidueVol(protein)

    Input: protein is a pure protein sequence.

    Output: result is a dict form containing 30 Geary Autocorrelation

    descriptors based on ResidueVol.
    ####################################################################################
    """

    result = CalculateEachGearyAuto(ProteinSequence, _ResidueVol, "_ResidueVol")
    return result


def CalculateGearyAutoSteric(ProteinSequence):
    """
    ####################################################################################
    Calculte the GearyAuto Autocorrelation descriptors based on

    Steric.

    Usage:

    result=CalculateGearyAutoSteric(protein)

    Input: protein is a pure protein sequence.

    Output: result is a dict form containing 30 Geary Autocorrelation

    descriptors based on Steric.
    ####################################################################################
    """

    result = CalculateEachGearyAuto(ProteinSequence, _Steric, "_Steric")
    return result


def CalculateGearyAutoMutability(ProteinSequence):
    """
    ####################################################################################
    Calculte the GearyAuto Autocorrelation descriptors based on

    Mutability.

    Usage:

    result=CalculateGearyAutoMutability(protein)

    Input: protein is a pure protein sequence.

    Output: result is a dict form containing 30 Geary Autocorrelation

    descriptors based on Mutability.
    ####################################################################################
    """

    result = CalculateEachGearyAuto(ProteinSequence, _Mutability, "_Mutability")
    return result


##################################################################################################


def CalculateNormalizedMoreauBrotoAutoTotal(ProteinSequence):
    """
    ####################################################################################
    A method used for computing normalized Moreau Broto autocorrelation descriptors based

    on 8 proterties of AADs.

    Usage:

    result=CalculateNormalizedMoreauBrotoAutoTotal(protein)

    Input: protein is a pure protein sequence.

    Output: result is a dict form containing 30*8=240 normalized Moreau Broto

    autocorrelation descriptors based on the given properties(i.e., _AAPropert).
    ####################################################################################
    """
    result = {}
    result.update(CalculateNormalizedMoreauBrotoAutoHydrophobicity(ProteinSequence))
    result.update(CalculateNormalizedMoreauBrotoAutoAvFlexibility(ProteinSequence))
    result.update(CalculateNormalizedMoreauBrotoAutoPolarizability(ProteinSequence))
    result.update(CalculateNormalizedMoreauBrotoAutoFreeEnergy(ProteinSequence))
    result.update(CalculateNormalizedMoreauBrotoAutoResidueASA(ProteinSequence))
    result.update(CalculateNormalizedMoreauBrotoAutoResidueVol(ProteinSequence))
    result.update(CalculateNormalizedMoreauBrotoAutoSteric(ProteinSequence))
    result.update(CalculateNormalizedMoreauBrotoAutoMutability(ProteinSequence))
    return result


def CalculateMoranAutoTotal(ProteinSequence):
    """
    ####################################################################################
    A method used for computing Moran autocorrelation descriptors based on 8 properties of AADs.

    Usage:

    result=CalculateMoranAutoTotal(protein)

    Input: protein is a pure protein sequence.

    Output: result is a dict form containing 30*8=240 Moran

    autocorrelation descriptors based on the given properties(i.e., _AAPropert).
    ####################################################################################
    """
    result = {}
    result.update(CalculateMoranAutoHydrophobicity(ProteinSequence))
    result.update(CalculateMoranAutoAvFlexibility(ProteinSequence))
    result.update(CalculateMoranAutoPolarizability(ProteinSequence))
    result.update(CalculateMoranAutoFreeEnergy(ProteinSequence))
    result.update(CalculateMoranAutoResidueASA(ProteinSequence))
    result.update(CalculateMoranAutoResidueVol(ProteinSequence))
    result.update(CalculateMoranAutoSteric(ProteinSequence))
    result.update(CalculateMoranAutoMutability(ProteinSequence))
    return result


def CalculateGearyAutoTotal(ProteinSequence):
    """
    ####################################################################################
    A method used for computing Geary autocorrelation descriptors based on 8 properties of AADs.

    Usage:

    result=CalculateGearyAutoTotal(protein)

    Input: protein is a pure protein sequence.

    Output: result is a dict form containing 30*8=240 Geary

    autocorrelation descriptors based on the given properties(i.e., _AAPropert).
    ####################################################################################
    """
    result = {}
    result.update(CalculateGearyAutoHydrophobicity(ProteinSequence))
    result.update(CalculateGearyAutoAvFlexibility(ProteinSequence))
    result.update(CalculateGearyAutoPolarizability(ProteinSequence))
    result.update(CalculateGearyAutoFreeEnergy(ProteinSequence))
    result.update(CalculateGearyAutoResidueASA(ProteinSequence))
    result.update(CalculateGearyAutoResidueVol(ProteinSequence))
    result.update(CalculateGearyAutoSteric(ProteinSequence))
    result.update(CalculateGearyAutoMutability(ProteinSequence))
    return result


##################################################################################################
def CalculateAutoTotal(ProteinSequence):
    """
    ####################################################################################
    A method used for computing all autocorrelation descriptors based on 8 properties of AADs.

    Usage:

    result=CalculateGearyAutoTotal(protein)

    Input: protein is a pure protein sequence.

    Output: result is a dict form containing 30*8*3=720 normalized Moreau Broto, Moran, and Geary

    autocorrelation descriptors based on the given properties(i.e., _AAPropert).
    ####################################################################################
    """
    result = {}
    result.update(CalculateNormalizedMoreauBrotoAutoTotal(ProteinSequence))
    result.update(CalculateMoranAutoTotal(ProteinSequence))
    result.update(CalculateGearyAutoTotal(ProteinSequence))
    return result


##################################################################################################
if __name__ == "__main__":
    protein = "ADGCGVGEGTGQGPMCNCMCMKWVYADEDAADLESDSFADEDASLESDSFPWSNQRVFCSFADEDAS"
    temp1 = CalculateNormalizedMoreauBrotoAuto(
        protein, AAProperty=_AAProperty, AAPropertyName=_AAPropertyName
    )
    # print temp1
    temp2 = CalculateMoranAutoMutability(protein)
    print(temp2)
    print(len(CalculateAutoTotal(protein)))



# -*- coding: utf-8 -*-
#  Copyright (c) 2016-2017, Zhijiang Yao, Jie Dong and Dongsheng Cao
#  All rights reserved.
#  This file is part of the PyBioMed.
#  The contents are covered by the terms of the BSD license
#  which is included in the file license.txt, found at the root
#  of the PyBioMed source tree.
"""
A class used for computing different types of protein descriptors!

You can freely use and distribute it. If you have any problem,

you could contact with us timely.

Authors: Zhijiang Yao and Dongsheng Cao.

Date: 2016.06.04

Email: gadsby@163.com

"""
# First party modules
from PyBioMed.PyProtein.AAComposition import (
    CalculateAAComposition,
    CalculateDipeptideComposition,
    GetSpectrumDict,
)
from PyBioMed.PyProtein.AAIndex import GetAAIndex1, GetAAIndex23
from PyBioMed.PyProtein.Autocorrelation import (
    CalculateEachGearyAuto,
    CalculateEachMoranAuto,
    CalculateEachNormalizedMoreauBrotoAuto,
    CalculateGearyAutoTotal,
    CalculateMoranAutoTotal,
    CalculateNormalizedMoreauBrotoAutoTotal,
)
from PyBioMed.PyProtein.ConjointTriad import CalculateConjointTriad
from PyBioMed.PyProtein.CTD import CalculateCTD
from PyBioMed.PyProtein.GetSubSeq import GetSubSequence
from PyBioMed.PyProtein.PseudoAAC import GetAPseudoAAC, GetPseudoAAC, _GetPseudoAAC
from PyBioMed.PyProtein.QuasiSequenceOrder import (
    GetQuasiSequenceOrder,
    GetQuasiSequenceOrderp,
    GetSequenceOrderCouplingNumberp,
    GetSequenceOrderCouplingNumberTotal,
)


class PyProtein:
    """
    This GetProDes class aims at collecting all descriptor calcualtion modules into a simple class.

    """

    AALetter = [
        "A",
        "R",
        "N",
        "D",
        "C",
        "E",
        "Q",
        "G",
        "H",
        "I",
        "L",
        "K",
        "M",
        "F",
        "P",
        "S",
        "T",
        "W",
        "Y",
        "V",
    ]

    Version = 1.0

    def __init__(self, ProteinSequence=""):
        """
        input a protein sequence
        """
        if len(ProteinSequence) == 0:
            print(
                "You must input a protein sequence when constructing a object. It is a string!"
            )
        else:
            self.ProteinSequence = ProteinSequence

    def GetAAComp(self):
        """
        amino acid compositon descriptors (20)

        Usage:

        result = GetAAComp()
        """
        res = CalculateAAComposition(self.ProteinSequence)
        return res

    def GetDPComp(self):
        """
        dipeptide composition descriptors (400)

        Usage:

        result = GetDPComp()
        """
        res = CalculateDipeptideComposition(self.ProteinSequence)
        return res

    def GetTPComp(self):
        """
        tri-peptide composition descriptors (8000)

        Usage:

        result = GetTPComp()
        """
        res = GetSpectrumDict(self.ProteinSequence)
        return res

    def GetMoreauBrotoAuto(self):
        """
        Normalized Moreau-Broto autocorrelation descriptors (240)

        Usage:

        result = GetMoreauBrotoAuto()
        """
        res = CalculateNormalizedMoreauBrotoAutoTotal(self.ProteinSequence)
        return res

    def GetMoranAuto(self):
        """
        Moran autocorrelation descriptors (240)

        Usage:

        result = GetMoranAuto()
        """
        res = CalculateMoranAutoTotal(self.ProteinSequence)
        return res

    def GetGearyAuto(self):
        """
        Geary autocorrelation descriptors (240)

        Usage:

        result = GetGearyAuto()
        """
        res = CalculateGearyAutoTotal(self.ProteinSequence)
        return res

    def GetCTD(self):
        """
        Composition Transition Distribution descriptors (147)

        Usage:

        result = GetCTD()
        """
        res = CalculateCTD(self.ProteinSequence)
        return res

    def GetPAAC(self, lamda=10, weight=0.05):
        """
        Type I Pseudo amino acid composition descriptors (default is 30)

        Usage:

        result = GetPAAC(lamda=10,weight=0.05)

        lamda factor reflects the rank of correlation and is a non-Negative integer, such as 15.

        Note that (1)lamda should NOT be larger than the length of input protein sequence;

        (2) lamda must be non-Negative integer, such as 0, 1, 2, ...; (3) when lamda =0, the

        output of PseAA server is the 20-D amino acid composition.

        weight factor is designed for the users to put weight on the additional PseAA components

        with respect to the conventional AA components. The user can select any value within the

        region from 0.05 to 0.7 for the weight factor.
        """
        res = _GetPseudoAAC(self.ProteinSequence, lamda=lamda, weight=weight)
        return res

    def GetPAACp(self, lamda=10, weight=0.05, AAP=[]):
        """
        Type I Pseudo amino acid composition descriptors for the given properties (default is 30)

        Usage:

        result = GetPAACp(lamda=10,weight=0.05,AAP=[])

        lamda factor reflects the rank of correlation and is a non-Negative integer, such as 15.

        Note that (1)lamda should NOT be larger than the length of input protein sequence;

        (2) lamda must be non-Negative integer, such as 0, 1, 2, ...; (3) when lamda =0, the

        output of PseAA server is the 20-D amino acid composition.

        weight factor is designed for the users to put weight on the additional PseAA components

        with respect to the conventional AA components. The user can select any value within the

        region from 0.05 to 0.7 for the weight factor.

        AAP is a list form containing the properties, each of which is a dict form.
        """
        res = GetPseudoAAC(self.ProteinSequence, lamda=lamda, weight=weight, AAP=AAP)
        return res

    def GetAPAAC(self, lamda=10, weight=0.5):
        """
        Amphiphilic (Type II) Pseudo amino acid composition descriptors

        default is 30

        Usage:

        result = GetAPAAC(lamda=10,weight=0.5)

        lamda factor reflects the rank of correlation and is a non-Negative integer, such as 15.

        Note that (1)lamda should NOT be larger than the length of input protein sequence;

        (2) lamda must be non-Negative integer, such as 0, 1, 2, ...; (3) when lamda =0, the

        output of PseAA server is the 20-D amino acid composition.

        weight factor is designed for the users to put weight on the additional PseAA components

        with respect to the conventional AA components. The user can select any value within the

        region from 0.05 to 0.7 for the weight factor.

        """
        res = GetAPseudoAAC(self.ProteinSequence, lamda=lamda, weight=weight)
        return res

    def GetSOCN(self, maxlag=45):
        """
        Sequence order coupling numbers  default is 45

        Usage:

        result = GetSOCN(maxlag=45)

        maxlag is the maximum lag and the length of the protein should be larger

        than maxlag. default is 45.
        """
        res = GetSequenceOrderCouplingNumberTotal(self.ProteinSequence, maxlag=maxlag)
        return res

    def GetSOCNp(self, maxlag=45, distancematrix={}):
        """
        Sequence order coupling numbers  default is 45

        Usage:

        result = GetSOCN(maxlag=45)

        maxlag is the maximum lag and the length of the protein should be larger

        than maxlag. default is 45.

        distancematrix is a dict form containing 400 distance values
        """
        res = GetSequenceOrderCouplingNumberp(
            self.ProteinSequence, maxlag=maxlag, distancematrix=distancematrix
        )
        return res

    def GetQSO(self, maxlag=30, weight=0.1):
        """
        Quasi sequence order descriptors  default is 50

        result = GetQSO(maxlag=30, weight=0.1)

        maxlag is the maximum lag and the length of the protein should be larger

        than maxlag. default is 45.
        """
        res = GetQuasiSequenceOrder(self.ProteinSequence, maxlag=maxlag, weight=weight)
        return res

    def GetQSOp(self, maxlag=30, weight=0.1, distancematrix={}):
        """
        Quasi sequence order descriptors  default is 50

        result = GetQSO(maxlag=30, weight=0.1)

        maxlag is the maximum lag and the length of the protein should be larger

        than maxlag. default is 45.

        distancematrix is a dict form containing 400 distance values
        """
        res = GetQuasiSequenceOrderp(
            self.ProteinSequence,
            maxlag=maxlag,
            weight=weight,
            distancematrix=distancematrix,
        )
        return res

    def GetMoreauBrotoAutop(self, AAP={}, AAPName="p"):
        """
        Normalized Moreau-Broto autocorrelation descriptors for the given property (30)

        Usage:

        result = GetMoreauBrotoAutop(AAP={},AAPName='p')

        AAP is a dict containing physicochemical properities of 20 amino acids
        """
        res = CalculateEachNormalizedMoreauBrotoAuto(
            self.ProteinSequence, AAP=AAP, AAPName=AAPName
        )
        return res

    def GetMoranAutop(self, AAP={}, AAPName="p"):
        """
        Moran autocorrelation descriptors for the given property (30)

        Usage:

        result = GetMoranAutop(AAP={},AAPName='p')

        AAP is a dict containing physicochemical properities of 20 amino acids
        """
        res = CalculateEachMoranAuto(self.ProteinSequence, AAP=AAP, AAPName=AAPName)
        return res

    def GetGearyAutop(self, AAP={}, AAPName="p"):
        """
        Geary autocorrelation descriptors for the given property (30)

        Usage:

        result = GetGearyAutop(AAP={},AAPName='p')

        AAP is a dict containing physicochemical properities of 20 amino acids
        """
        res = CalculateEachGearyAuto(self.ProteinSequence, AAP=AAP, AAPName=AAPName)
        return res

    def GetSubSeq(self, ToAA="S", window=3):
        """
        obtain the sub sequences wit length 2*window+1, whose central point is ToAA

        Usage:

        result = GetSubSeq(ToAA='S',window=3)

        ToAA is the central (query point) amino acid in the sub-sequence.

        window is the span.
        """
        res = GetSubSequence(self.ProteinSequence, ToAA=ToAA, window=window)
        return res

    def GetTriad(self):
        """
        Calculate the conjoint triad features from protein sequence.

        Useage:

        res = GetTriad()

        Output is a dict form containing all 343 conjoint triad features.
        """
        res = CalculateConjointTriad(self.ProteinSequence)
        return res

    def GetALL(self):
        """
        Calcualte all descriptors except tri-peptide descriptors
        """
        res = {}
        res.update(self.GetAAComp())
        res.update(self.GetDPComp())
        res.update(self.GetTPComp())
        res.update(self.GetMoreauBrotoAuto())
        res.update(self.GetMoranAuto())
        res.update(self.GetGearyAuto())
        res.update(self.GetCTD())
        res.update(self.GetPAAC())
        res.update(self.GetAPAAC())
        res.update(self.GetSOCN())
        res.update(self.GetQSO())
        res.update(self.GetTriad())
        return res

    def GetAAindex1(self, name, path="."):
        """
        Get the amino acid property values from aaindex1

        Usage:

        result=GetAAIndex1(name)

        Input: name is the name of amino acid property (e.g., KRIW790103)

        Output: result is a dict form containing the properties of 20 amino acids
        """

        return GetAAIndex1(name, path=path)

    def GetAAindex23(self, name, path="."):
        """
        Get the amino acid property values from aaindex2 and aaindex3

        Usage:

        result=GetAAIndex23(name)

        Input: name is the name of amino acid property (e.g.,TANS760101,GRAR740104)

        Output: result is a dict form containing the properties of 400 amino acid pairs
        """
        return GetAAIndex23(name, path=path)


#####################################################################################################
if __name__ == "__main__":
    import os
    from Autocorrelation import _Steric
    from PseudoAAC import _Hydrophobicity, _hydrophilicity, _residuemass

    protein = "ADGCGVGEGTGQGPMCNCMCMKWVYADEDAADLESDSFADEDASLESDSFPWSNQRVFCSFADEDAS"
    cds = PyProtein(protein)

    print(cds.GetAAComp())
    print(cds.GetDPComp())
    print(cds.GetTPComp())
    print(cds.GetCTD())
    print(cds.GetPAAC(lamda=5))
    print(cds.GetALL())
    print(cds.GetMoreauBrotoAutop(AAP=_Steric, AAPName="Steric"))
    print(cds.GetMoranAutop(AAP=_Steric, AAPName="Steric"))
    print(cds.GetGearyAutop(AAP=_Steric, AAPName="Steric"))

    print(cds.GetPAACp(lamda=5, weight=0.05, AAP=[_Hydrophobicity, _hydrophilicity]))

    print(cds.GetSubSeq(ToAA="D", window=5))
    print(cds.GetTriad())
    proper = cds.GetAAindex23("GRAR740104")
    print(cds.GetAAindex1("KRIW790103"))

    print(cds.GetQSOp(maxlag=30, weight=0.1, distancematrix=proper))
    print(cds.GetSOCNp(maxlag=30, distancematrix=proper))



# -*- coding: utf-8 -*-
#  Copyright (c) 2016-2017, Zhijiang Yao, Jie Dong and Dongsheng Cao
#  All rights reserved.
#  This file is part of the PyBioMed.
#  The contents are covered by the terms of the BSD license
#  which is included in the file license.txt, found at the root
#  of the PyBioMed source tree.
"""
This module is used for obtaining the properties of amino acids or their pairs

from the aaindex database. You can freely use and distribute it. If you hava

any problem, you could contact with us timely!

Authors: Zhijiang Yao and Dongsheng Cao.

Date: 2016.06.04

Email: gadsby@163.com

"""

from __future__ import print_function

# Core Library modules
import os
import string
import sys

AALetter = [
    "A",
    "R",
    "N",
    "D",
    "C",
    "E",
    "Q",
    "G",
    "H",
    "I",
    "L",
    "K",
    "M",
    "F",
    "P",
    "S",
    "T",
    "W",
    "Y",
    "V",
]
_aaindex = dict()


#####################################################################################################
class Record:
    """
    Amino acid index (AAindex) Record
    """

    aakeys = "ARNDCQEGHILKMFPSTWYV"

    def __init__(self):
        self.key = None
        self.desc = ""
        self.ref = ""
        self.authors = ""
        self.title = ""
        self.journal = ""
        self.correlated = dict()
        self.index = dict()
        self.comment = ""

    def extend(self, row):
        i = len(self.index)
        for x in row:
            self.index[self.aakeys[i]] = x
            i += 1

    def get(self, aai, aaj=None, d=None):
        assert aaj is None
        return self.index.get(aai, d)

    def __getitem__(self, aai):
        return self.get(aai)

    def median(self):
        x = sorted(filter(None, self.index.values()))
        half = len(x) / 2
        if len(x) % 2 == 1:
            return x[half]
        return (x[half - 1] + x[half]) / 2.0

    def __str__(self):
        desc = self.desc.replace("\n", " ").strip()
        return "%s(%s: %s)" % (self.__class__.__name__, self.key, desc)


#####################################################################################################
class MatrixRecord(Record):
    """
    Matrix record for mutation matrices or pair-wise contact potentials
    """

    def __init__(self):
        Record.__init__(self)
        self.index = []
        self.rows = dict()
        self.cols = dict()

    def extend(self, row):
        self.index.append(row)

    def _get(self, aai, aaj):
        i = self.rows[aai]
        j = self.cols[aaj]
        return self.index[i][j]

    def get(self, aai, aaj, d=None):
        try:
            return self._get(aai, aaj)
        except:
            pass
        try:
            return self._get(aaj, aai)
        except:
            return d

    def __getitem__(self, aaij):
        return self.get(aaij[0], aaij[1])

    def median(self):
        x = []
        for y in self.index:
            x.extend(filter(None, y))
        x.sort()
        if len(x) % 2 == 1:
            return x[len(x) / 2]
        return sum(x[len(x) / 2 - 1 : len(x) / 2 + 1]) / 2.0


#####################################################################################################
def search(pattern, searchtitle=True, casesensitive=False):
    """
    Search for pattern in description and title (optional) of all records and
    return matched records as list. By default search case insensitive.
    """
    whatcase = lambda i: i
    if not casesensitive:
        pattern = pattern.lower()
        whatcase = lambda i: i.lower()
    matches = []
    for record in _aaindex.itervalues():
        if (
            pattern in whatcase(record.desc)
            or searchtitle
            and pattern in whatcase(record.title)
        ):
            matches.append(record)
    return matches


#####################################################################################################
def grep(pattern):
    """
    Search for pattern in title and description of all records (case
    insensitive) and print results on standard output.

    """
    for record in search(pattern):
        print(record)


#####################################################################################################
def get(key):
    """
    Get record for key
    """
    if len(_aaindex) == 0:
        init()
    return _aaindex[key]


#####################################################################################################
def _float_or_None(x):
    if x == "NA" or x == "-":
        return None
    return float(x)


#####################################################################################################
def init(path=None, index="123"):
    """
    Read in the aaindex files. You need to run this (once) before you can
    access any records. If the files are not within the current directory,
    you need to specify the correct directory path. By default all three
    aaindex files are read in.
    """
    index = str(index)
    if path is None:
        for path in [os.path.split(__file__)[0], "."]:
            if os.path.exists(os.path.join(path, "aaindex" + index[0])):
                break
        print("path =", path, file=sys.stderr)
    if "1" in index:
        _parse(path + "/aaindex1", Record)
    if "2" in index:
        _parse(path + "/aaindex2", MatrixRecord)
    if "3" in index:
        _parse(path + "/aaindex3", MatrixRecord)


#####################################################################################################
def init_from_file(filename, type=Record):
    _parse(filename, type)


#####################################################################################################
def _parse(filename, rec, quiet=True):
    """
    Parse aaindex input file. `rec` must be `Record` for aaindex1 and
    `MarixRecord` for aaindex2 and aaindex3.
    """
    if not os.path.exists(filename):
        import urllib

        url = (
            "ftp://ftp.genome.jp/pub/db/community/aaindex/" + os.path.split(filename)[1]
        )
        # 		print 'Downloading "%s"' % (url)
        filename = urllib.urlretrieve(url, filename)[0]
    # 		print 'Saved to "%s"' % (filename)
    f = open(filename)

    current = rec()
    lastkey = None
    for line in f:
        key = line[0:2]
        if key[0] == " ":
            key = lastkey
        if key == "//":
            _aaindex[current.key] = current
            current = rec()
        elif key == "H ":
            current.key = line[2:].strip()
        elif key == "R ":
            current.ref += line[2:]
        elif key == "D ":
            current.desc += line[2:]
        elif key == "A ":
            current.authors += line[2:]
        elif key == "T ":
            current.title += line[2:]
        elif key == "J ":
            current.journal += line[2:]
        elif key == "* ":
            current.comment += line[2:]
        elif key == "C ":
            a = line[2:].split()
            for i in range(0, len(a), 2):
                current.correlated[a[i]] = float(a[i + 1])
        elif key == "I ":
            a = line[1:].split()
            if a[0] != "A/L":
                current.extend(map(_float_or_None, a))
            elif list(Record.aakeys) != [i[0] for i in a] + [i[-1] for i in a]:
                print("Warning: wrong amino acid sequence for", current.key)
            else:
                try:
                    assert list(Record.aakeys[:10]) == [i[0] for i in a]
                    assert list(Record.aakeys[10:]) == [i[2] for i in a]
                except:
                    print("Warning: wrong amino acid sequence for", current.key)
        elif key == "M ":
            a = line[2:].split()
            if a[0] == "rows":
                if a[4] == "rows":
                    a.pop(4)
                assert a[3] == "cols" and len(a) == 6
                i = 0
                for aa in a[2]:
                    current.rows[aa] = i
                    i += 1
                i = 0
                for aa in a[5]:
                    current.cols[aa] = i
                    i += 1
            else:
                current.extend(map(_float_or_None, a))
        elif not quiet:
            print('Warning: line starts with "%s"' % (key))
        lastkey = key
    f.close()


#####################################################################################################
def GetAAIndex1(name, path="."):
    """
    Get the amino acid property values from aaindex1

    Usage:

    result=GetAAIndex1(name)

    Input: name is the name of amino acid property (e.g., KRIW790103)

    Output: result is a dict form containing the properties of 20 amino acids
    """

    init(path=path)
    name = str(name)
    temp = get(string.strip(name))
    res = {}
    for i in AALetter:
        res[i] = temp.get(i)
    return res


#####################################################################################################
def GetAAIndex23(name, path="."):
    """
    Get the amino acid property values from aaindex2 and aaindex3

    Usage:

    result=GetAAIndex23(name)

    Input: name is the name of amino acid property (e.g.,TANS760101,GRAR740104)

    Output: result is a dict form containing the properties of 400 amino acid pairs
    """
    init(path=path)
    name = str(name)
    temp = get(string.strip(name))
    res = {}
    for i in AALetter:
        for j in AALetter:
            res[i + j] = temp.get(i, j)
    return res


#####################################################################################################

if __name__ == "__main__":
    # 	init(path='.')

    # 	grep('volume')
    # 	x = get('KRIW790103')
    # 	print x
    # 	print x.get('W')
    temp1 = GetAAIndex1("KRIW790103")
    print(len(temp1))

    temp2 = GetAAIndex23("TANS760101")
    print(len(temp2))
    temp2 = GetAAIndex23("GRAR740104")
    print(len(temp2))



# -*- coding: utf-8 -*-
#  Copyright (c) 2016-2017, Zhijiang Yao, Jie Dong and Dongsheng Cao
#  All rights reserved.
#  This file is part of the PyBioMed.
#  The contents are covered by the terms of the BSD license
#  which is included in the file license.txt, found at the root
#  of the PyBioMed source tree.
"""
#########################################################################################

Instead of using the conventional 20-D amino acid composition to represent the sample

of a protein, Prof. Kuo-Chen Chou proposed the pseudo amino acid (PseAA) composition

in order for inluding the sequence-order information. Based on the concept of Chou's

pseudo amino acid composition, the server PseAA was designed in a flexible way, allowing

users to generate various kinds of pseudo amino acid composition for a given protein

sequence by selecting different parameters and their combinations. This module aims at

computing two types of PseAA descriptors: Type I and Type II.

You can freely use and distribute it. If you have any problem, you could contact

with us timely.

References:

[1]: Kuo-Chen Chou. Prediction of Protein Cellular Attributes Using Pseudo-Amino Acid

Composition. PROTEINS: Structure, Function, and Genetics, 2001, 43: 246-255.

[2]: http://www.csbio.sjtu.edu.cn/bioinf/PseAAC/

[3]: http://www.csbio.sjtu.edu.cn/bioinf/PseAAC/type2.htm

[4]: Kuo-Chen Chou. Using amphiphilic pseudo amino acid composition to predict enzyme

subfamily classes. Bioinformatics, 2005,21,10-19.

Authors: Zhijiang Yao and Dongsheng Cao.

Date: 2016.06.04

Email: gadsby@163.com

#########################################################################################
"""

# Core Library modules
import math
import string

# import scipy


AALetter = [
    "A",
    "R",
    "N",
    "D",
    "C",
    "E",
    "Q",
    "G",
    "H",
    "I",
    "L",
    "K",
    "M",
    "F",
    "P",
    "S",
    "T",
    "W",
    "Y",
    "V",
]

_Hydrophobicity = {
    "A": 0.62,
    "R": -2.53,
    "N": -0.78,
    "D": -0.90,
    "C": 0.29,
    "Q": -0.85,
    "E": -0.74,
    "G": 0.48,
    "H": -0.40,
    "I": 1.38,
    "L": 1.06,
    "K": -1.50,
    "M": 0.64,
    "F": 1.19,
    "P": 0.12,
    "S": -0.18,
    "T": -0.05,
    "W": 0.81,
    "Y": 0.26,
    "V": 1.08,
}

_hydrophilicity = {
    "A": -0.5,
    "R": 3.0,
    "N": 0.2,
    "D": 3.0,
    "C": -1.0,
    "Q": 0.2,
    "E": 3.0,
    "G": 0.0,
    "H": -0.5,
    "I": -1.8,
    "L": -1.8,
    "K": 3.0,
    "M": -1.3,
    "F": -2.5,
    "P": 0.0,
    "S": 0.3,
    "T": -0.4,
    "W": -3.4,
    "Y": -2.3,
    "V": -1.5,
}

_residuemass = {
    "A": 15.0,
    "R": 101.0,
    "N": 58.0,
    "D": 59.0,
    "C": 47.0,
    "Q": 72.0,
    "E": 73.0,
    "G": 1.000,
    "H": 82.0,
    "I": 57.0,
    "L": 57.0,
    "K": 73.0,
    "M": 75.0,
    "F": 91.0,
    "P": 42.0,
    "S": 31.0,
    "T": 45.0,
    "W": 130.0,
    "Y": 107.0,
    "V": 43.0,
}

_pK1 = {
    "A": 2.35,
    "C": 1.71,
    "D": 1.88,
    "E": 2.19,
    "F": 2.58,
    "G": 2.34,
    "H": 1.78,
    "I": 2.32,
    "K": 2.20,
    "L": 2.36,
    "M": 2.28,
    "N": 2.18,
    "P": 1.99,
    "Q": 2.17,
    "R": 2.18,
    "S": 2.21,
    "T": 2.15,
    "V": 2.29,
    "W": 2.38,
    "Y": 2.20,
}

_pK2 = {
    "A": 9.87,
    "C": 10.78,
    "D": 9.60,
    "E": 9.67,
    "F": 9.24,
    "G": 9.60,
    "H": 8.97,
    "I": 9.76,
    "K": 8.90,
    "L": 9.60,
    "M": 9.21,
    "N": 9.09,
    "P": 10.6,
    "Q": 9.13,
    "R": 9.09,
    "S": 9.15,
    "T": 9.12,
    "V": 9.74,
    "W": 9.39,
    "Y": 9.11,
}

_pI = {
    "A": 6.11,
    "C": 5.02,
    "D": 2.98,
    "E": 3.08,
    "F": 5.91,
    "G": 6.06,
    "H": 7.64,
    "I": 6.04,
    "K": 9.47,
    "L": 6.04,
    "M": 5.74,
    "N": 10.76,
    "P": 6.30,
    "Q": 5.65,
    "R": 10.76,
    "S": 5.68,
    "T": 5.60,
    "V": 6.02,
    "W": 5.88,
    "Y": 5.63,
}


#############################################################################################


def _mean(listvalue):
    """
    ########################################################################################
    The mean value of the list data.

    Usage:

    result=_mean(listvalue)
    ########################################################################################
    """
    return sum(listvalue) / len(listvalue)


##############################################################################################
def _std(listvalue, ddof=1):
    """
    ########################################################################################
    The standard deviation of the list data.

    Usage:

    result=_std(listvalue)
    ########################################################################################
    """
    mean = _mean(listvalue)
    temp = [math.pow(i - mean, 2) for i in listvalue]
    res = math.sqrt(sum(temp) / (len(listvalue) - ddof))
    return res


##############################################################################################
def NormalizeEachAAP(AAP):
    """
    ########################################################################################
    All of the amino acid indices are centralized and

    standardized before the calculation.

    Usage:

    result=NormalizeEachAAP(AAP)

    Input: AAP is a dict form containing the properties of 20 amino acids.

    Output: result is the a dict form containing the normalized properties

    of 20 amino acids.
    ########################################################################################
    """
    if len(AAP.values()) != 20:
        print("You can not input the correct number of properities of Amino acids!")
    else:
        Result = {}
        for i, j in AAP.items():
            Result[i] = (j - _mean(AAP.values())) / _std(AAP.values(), ddof=0)

    return Result


#############################################################################################
#############################################################################################
##################################Type I descriptors#########################################
####################### Pseudo-Amino Acid Composition descriptors############################
#############################################################################################
#############################################################################################
def _GetCorrelationFunction(
    Ri="S", Rj="D", AAP=[_Hydrophobicity, _hydrophilicity, _residuemass]
):
    """
    ########################################################################################
    Computing the correlation between two given amino acids using the above three

    properties.

    Usage:

    result=_GetCorrelationFunction(Ri,Rj)

    Input: Ri and Rj are the amino acids, respectively.

    Output: result is the correlation value between two amino acids.
    ########################################################################################
    """
    Hydrophobicity = NormalizeEachAAP(AAP[0])
    hydrophilicity = NormalizeEachAAP(AAP[1])
    residuemass = NormalizeEachAAP(AAP[2])
    theta1 = math.pow(Hydrophobicity[Ri] - Hydrophobicity[Rj], 2)
    theta2 = math.pow(hydrophilicity[Ri] - hydrophilicity[Rj], 2)
    theta3 = math.pow(residuemass[Ri] - residuemass[Rj], 2)
    theta = round((theta1 + theta2 + theta3) / 3.0, 3)
    return theta


#############################################################################################


def _GetSequenceOrderCorrelationFactor(ProteinSequence, k=1):
    """
    ########################################################################################
    Computing the Sequence order correlation factor with gap equal to k based on

    [_Hydrophobicity,_hydrophilicity,_residuemass].

    Usage:

    result=_GetSequenceOrderCorrelationFactor(protein,k)

    Input: protein is a pure protein sequence.

    k is the gap.

    Output: result is the correlation factor value with the gap equal to k.
    ########################################################################################
    """
    LengthSequence = len(ProteinSequence)
    res = []
    for i in range(LengthSequence - k):
        AA1 = ProteinSequence[i]
        AA2 = ProteinSequence[i + k]
        res.append(_GetCorrelationFunction(AA1, AA2))
    result = round(sum(res) / (LengthSequence - k), 3)
    return result


#############################################################################################


def GetAAComposition(ProteinSequence):
    """
    ########################################################################################
    Calculate the composition of Amino acids

    for a given protein sequence.

    Usage:

    result=CalculateAAComposition(protein)

    Input: protein is a pure protein sequence.

    Output: result is a dict form containing the composition of

    20 amino acids.
    ########################################################################################
    """
    LengthSequence = len(ProteinSequence)
    Result = {}
    for i in AALetter:
        Result[i] = round(float(ProteinSequence.count(i)) / LengthSequence * 100, 3)
    return Result


#############################################################################################
def _GetPseudoAAC1(ProteinSequence, lamda=10, weight=0.05):
    """
    #######################################################################################
    Computing the first 20 of type I pseudo-amino acid compostion descriptors based on

    [_Hydrophobicity,_hydrophilicity,_residuemass].
    ########################################################################################
    """
    rightpart = 0.0
    for i in range(lamda):
        rightpart = rightpart + _GetSequenceOrderCorrelationFactor(
            ProteinSequence, k=i + 1
        )
    AAC = GetAAComposition(ProteinSequence)

    result = {}
    temp = 1 + weight * rightpart
    for index, i in enumerate(AALetter):
        result["PAAC" + str(index + 1)] = round(AAC[i] / temp, 3)

    return result


#############################################################################################
def _GetPseudoAAC2(ProteinSequence, lamda=10, weight=0.05):
    """
    ########################################################################################
    Computing the last lamda of type I pseudo-amino acid compostion descriptors based on

    [_Hydrophobicity,_hydrophilicity,_residuemass].
    ########################################################################################
    """
    rightpart = []
    for i in range(lamda):
        rightpart.append(_GetSequenceOrderCorrelationFactor(ProteinSequence, k=i + 1))

    result = {}
    temp = 1 + weight * sum(rightpart)
    for index in range(20, 20 + lamda):
        result["PAAC" + str(index + 1)] = round(
            weight * rightpart[index - 20] / temp * 100, 3
        )

    return result


#############################################################################################


def _GetPseudoAAC(ProteinSequence, lamda=10, weight=0.05):
    """
    #######################################################################################
    Computing all of type I pseudo-amino acid compostion descriptors based on three given

    properties. Note that the number of PAAC strongly depends on the lamda value. if lamda

    = 20, we can obtain 20+20=40 PAAC descriptors. The size of these values depends on the

    choice of lamda and weight simultaneously.

    AAP=[_Hydrophobicity,_hydrophilicity,_residuemass]

    Usage:

    result=_GetAPseudoAAC(protein,lamda,weight)

    Input: protein is a pure protein sequence.

    lamda factor reflects the rank of correlation and is a non-Negative integer, such as 15.

    Note that (1)lamda should NOT be larger than the length of input protein sequence;

    (2) lamda must be non-Negative integer, such as 0, 1, 2, ...; (3) when lamda =0, the

    output of PseAA server is the 20-D amino acid composition.

    weight factor is designed for the users to put weight on the additional PseAA components

    with respect to the conventional AA components. The user can select any value within the

    region from 0.05 to 0.7 for the weight factor.

    Output: result is a dict form containing calculated 20+lamda PAAC descriptors.
    ########################################################################################
    """
    res = {}
    res.update(_GetPseudoAAC1(ProteinSequence, lamda=lamda, weight=weight))
    res.update(_GetPseudoAAC2(ProteinSequence, lamda=lamda, weight=weight))
    return res


#############################################################################################
##################################Type II descriptors########################################
###############Amphiphilic Pseudo-Amino Acid Composition descriptors#########################
#############################################################################################
#############################################################################################
def _GetCorrelationFunctionForAPAAC(
    Ri="S", Rj="D", AAP=[_Hydrophobicity, _hydrophilicity]
):
    """
    ########################################################################################
    Computing the correlation between two given amino acids using the above two

    properties for APAAC (type II PseAAC).

    Usage:

    result=_GetCorrelationFunctionForAPAAC(Ri,Rj)

    Input: Ri and Rj are the amino acids, respectively.

    Output: result is the correlation value between two amino acids.
    ########################################################################################
    """
    Hydrophobicity = NormalizeEachAAP(AAP[0])
    hydrophilicity = NormalizeEachAAP(AAP[1])
    theta1 = round(Hydrophobicity[Ri] * Hydrophobicity[Rj], 3)
    theta2 = round(hydrophilicity[Ri] * hydrophilicity[Rj], 3)

    return theta1, theta2


#############################################################################################
def GetSequenceOrderCorrelationFactorForAPAAC(ProteinSequence, k=1):
    """
    ########################################################################################
    Computing the Sequence order correlation factor with gap equal to k based on

    [_Hydrophobicity,_hydrophilicity] for APAAC (type II PseAAC) .

    Usage:

    result=GetSequenceOrderCorrelationFactorForAPAAC(protein,k)

    Input: protein is a pure protein sequence.

    k is the gap.

    Output: result is the correlation factor value with the gap equal to k.
    ########################################################################################
    """
    LengthSequence = len(ProteinSequence)
    resHydrophobicity = []
    reshydrophilicity = []
    for i in range(LengthSequence - k):
        AA1 = ProteinSequence[i]
        AA2 = ProteinSequence[i + k]
        temp = _GetCorrelationFunctionForAPAAC(AA1, AA2)
        resHydrophobicity.append(temp[0])
        reshydrophilicity.append(temp[1])
    result = []
    result.append(round(sum(resHydrophobicity) / (LengthSequence - k), 3))
    result.append(round(sum(reshydrophilicity) / (LengthSequence - k), 3))
    return result


#############################################################################################
def GetAPseudoAAC1(ProteinSequence, lamda=30, weight=0.5):
    """
    ########################################################################################
    Computing the first 20 of type II pseudo-amino acid compostion descriptors based on

    [_Hydrophobicity,_hydrophilicity].
    ########################################################################################
    """
    rightpart = 0.0
    for i in range(lamda):
        rightpart = rightpart + sum(
            GetSequenceOrderCorrelationFactorForAPAAC(ProteinSequence, k=i + 1)
        )
    AAC = GetAAComposition(ProteinSequence)

    result = {}
    temp = 1 + weight * rightpart
    for index, i in enumerate(AALetter):
        result["APAAC" + str(index + 1)] = round(AAC[i] / temp, 3)

    return result


#############################################################################################
def GetAPseudoAAC2(ProteinSequence, lamda=30, weight=0.5):
    """
    #######################################################################################
    Computing the last lamda of type II pseudo-amino acid compostion descriptors based on

    [_Hydrophobicity,_hydrophilicity].
    #######################################################################################
    """
    rightpart = []
    for i in range(lamda):
        temp = GetSequenceOrderCorrelationFactorForAPAAC(ProteinSequence, k=i + 1)
        rightpart.append(temp[0])
        rightpart.append(temp[1])

    result = {}
    temp = 1 + weight * sum(rightpart)
    for index in range(20, 20 + 2 * lamda):
        result["PAAC" + str(index + 1)] = round(
            weight * rightpart[index - 20] / temp * 100, 3
        )

    return result


#############################################################################################
def GetAPseudoAAC(ProteinSequence, lamda=30, weight=0.5):
    """
    #######################################################################################
    Computing all of type II pseudo-amino acid compostion descriptors based on the given

    properties. Note that the number of PAAC strongly depends on the lamda value. if lamda

    = 20, we can obtain 20+20=40 PAAC descriptors. The size of these values depends on the

    choice of lamda and weight simultaneously.

    Usage:

    result=GetAPseudoAAC(protein,lamda,weight)

    Input: protein is a pure protein sequence.

    lamda factor reflects the rank of correlation and is a non-Negative integer, such as 15.

    Note that (1)lamda should NOT be larger than the length of input protein sequence;

    (2) lamda must be non-Negative integer, such as 0, 1, 2, ...; (3) when lamda =0, the

    output of PseAA server is the 20-D amino acid composition.

    weight factor is designed for the users to put weight on the additional PseAA components

    with respect to the conventional AA components. The user can select any value within the

    region from 0.05 to 0.7 for the weight factor.

    Output: result is a dict form containing calculated 20+lamda PAAC descriptors.
    #######################################################################################
    """
    res = {}
    res.update(GetAPseudoAAC1(ProteinSequence, lamda=lamda, weight=weight))
    res.update(GetAPseudoAAC2(ProteinSequence, lamda=lamda, weight=weight))
    return res


#############################################################################################
#############################################################################################
##################################Type I descriptors#########################################
####################### Pseudo-Amino Acid Composition descriptors############################
#############################based on different properties###################################
#############################################################################################
#############################################################################################
def GetCorrelationFunction(Ri="S", Rj="D", AAP=[]):
    """
    ########################################################################################
    Computing the correlation between two given amino acids using the given

    properties.

    Usage:

    result=GetCorrelationFunction(Ri,Rj,AAP)

    Input: Ri and Rj are the amino acids, respectively.

    AAP is a list form containing the properties, each of which is a dict form.

    Output: result is the correlation value between two amino acids.
    ########################################################################################
    """
    NumAAP = len(AAP)
    theta = 0.0
    for i in range(NumAAP):
        temp = NormalizeEachAAP(AAP[i])
        theta = theta + math.pow(temp[Ri] - temp[Rj], 2)
    result = round(theta / NumAAP, 3)
    return result


#############################################################################################
def GetSequenceOrderCorrelationFactor(ProteinSequence, k=1, AAP=[]):
    """
    ########################################################################################
    Computing the Sequence order correlation factor with gap equal to k based on

    the given properities.

    Usage:

    result=GetSequenceOrderCorrelationFactor(protein,k,AAP)

    Input: protein is a pure protein sequence.

    k is the gap.

    AAP is a list form containing the properties, each of which is a dict form.

    Output: result is the correlation factor value with the gap equal to k.
    ########################################################################################
    """
    LengthSequence = len(ProteinSequence)
    res = []
    for i in range(LengthSequence - k):
        AA1 = ProteinSequence[i]
        AA2 = ProteinSequence[i + k]
        res.append(GetCorrelationFunction(AA1, AA2, AAP))
    result = round(sum(res) / (LengthSequence - k), 3)
    return result


#############################################################################################
def GetPseudoAAC1(ProteinSequence, lamda=30, weight=0.05, AAP=[]):
    """
    #######################################################################################
    Computing the first 20 of type I pseudo-amino acid compostion descriptors based on the given

    properties.
    ########################################################################################
    """
    rightpart = 0.0
    for i in range(lamda):
        rightpart = rightpart + GetSequenceOrderCorrelationFactor(
            ProteinSequence, i + 1, AAP
        )
    AAC = GetAAComposition(ProteinSequence)

    result = {}
    temp = 1 + weight * rightpart
    for index, i in enumerate(AALetter):
        result["PAAC" + str(index + 1)] = round(AAC[i] / temp, 3)

    return result


#############################################################################################
def GetPseudoAAC2(ProteinSequence, lamda=30, weight=0.05, AAP=[]):
    """
    #######################################################################################
    Computing the last lamda of type I pseudo-amino acid compostion descriptors based on the given

    properties.
    ########################################################################################
    """
    rightpart = []
    for i in range(lamda):
        rightpart.append(GetSequenceOrderCorrelationFactor(ProteinSequence, i + 1, AAP))

    result = {}
    temp = 1 + weight * sum(rightpart)
    for index in range(20, 20 + lamda):
        result["PAAC" + str(index + 1)] = round(
            weight * rightpart[index - 20] / temp * 100, 3
        )

    return result


#############################################################################################


def GetPseudoAAC(ProteinSequence, lamda=30, weight=0.05, AAP=[]):
    """
    #######################################################################################
    Computing all of type I pseudo-amino acid compostion descriptors based on the given

    properties. Note that the number of PAAC strongly depends on the lamda value. if lamda

    = 20, we can obtain 20+20=40 PAAC descriptors. The size of these values depends on the

    choice of lamda and weight simultaneously. You must specify some properties into AAP.

    Usage:

    result=GetPseudoAAC(protein,lamda,weight)

    Input: protein is a pure protein sequence.

    lamda factor reflects the rank of correlation and is a non-Negative integer, such as 15.

    Note that (1)lamda should NOT be larger than the length of input protein sequence;

    (2) lamda must be non-Negative integer, such as 0, 1, 2, ...; (3) when lamda =0, the

    output of PseAA server is the 20-D amino acid composition.

    weight factor is designed for the users to put weight on the additional PseAA components

    with respect to the conventional AA components. The user can select any value within the

    region from 0.05 to 0.7 for the weight factor.

    AAP is a list form containing the properties, each of which is a dict form.

    Output: result is a dict form containing calculated 20+lamda PAAC descriptors.
    ########################################################################################
    """
    res = {}
    res.update(GetPseudoAAC1(ProteinSequence, lamda, weight, AAP))
    res.update(GetPseudoAAC2(ProteinSequence, lamda, weight, AAP))
    return res


#############################################################################################

if __name__ == "__main__":
    import string

    protein = "MTDRARLRLHDTAAGVVRDFVPLRPGHVSIYLCGATVQGLPHIGHVRSGVAFDILRRWLL\
ARGYDVAFIRNVTDIEDKILAKAAAAGRPWWEWAATHERAFTAAYDALDVLPPSAEPRAT\
GHITQMIEMIERLIQAGHAYTGGGDVYFDVLSYPEYGQLSGHKIDDVHQGEGVAAGKRDQ\
RDFTLWKGEKPGEPSWPTPWGRGRPGWHLECSAMARSYLGPEFDIHCGGMDLVFPHHENE\
IAQSRAAGDGFARYWLHNGWVTMGGEKMSKSLGNVLSMPAMLQRVRPAELRYYLGSAHYR\
SMLEFSETAMQDAVKAYVGLEDFLHRVRTRVGAVCPGDPTPRFAEALDDDLSVPIALAEI\
HHVRAEGNRALDAGDHDGALRSASAIRAMMGILGCDPLDQRWESRDETSAALAAVDVLVQ\
AELQNREKAREQRNWALADEIRGRLKRAGIEVTDTADGPQWSLLGGDTK"
    protein = string.strip(protein)
    # 	temp=_GetCorrelationFunction('S','D')
    # 	print temp
    #
    # 	print _GetSequenceOrderCorrelationFactor(protein,k=4)
    #
    # 	PAAC1=_GetPseudoAAC1(protein,lamda=4)
    # 	for i in PAAC1:
    # 		print i, PAAC1[i]
    # 	PAAC2=_GetPseudoAAC2(protein,lamda=4)
    # 	for i in PAAC2:
    # 		print i, PAAC2[i]
    # 	print len(PAAC1)
    # 	print _GetSequenceOrderCorrelationFactorForAPAAC(protein,k=1)
    # 	APAAC1=_GetAPseudoAAC1(protein,lamda=4)
    # 	for i in APAAC1:
    # 		print i, APAAC1[i]

    # 	APAAC2=GetAPseudoAAC2(protein,lamda=4)
    # 	for i in APAAC2:
    # 		print i, APAAC2[i]
    # 	APAAC=GetAPseudoAAC(protein,lamda=4)
    #
    # 	for i in APAAC:
    # 		print i, APAAC[i]

    PAAC = GetPseudoAAC(protein, lamda=5, AAP=[_Hydrophobicity, _hydrophilicity])

    for i in PAAC:
        print(i, PAAC[i])





################################################################################
#################                    Model                     #################
################################################################################

from sklearn.neighbors import KNeighborsRegressor
from sklearn.svm import SVR
from sklearn.linear_model import Lasso, LinearRegression
from sklearn.tree import DecisionTreeRegressor
from sklearn.ensemble import RandomForestRegressor, AdaBoostRegressor, BaggingRegressor
from sklearn.cross_decomposition import PLSRegression
from sklearn.preprocessing import StandardScaler, MinMaxScaler
from sklearn.model_selection import GridSearchCV, train_test_split
from sklearn.metrics import SCORERS
from sklearn.exceptions import UndefinedMetricWarning
from difflib import get_close_matches
import numpy as np
import inspect
import os
import pickle

from .evaluate import Evaluate

class Model():
    """
    Class for building, fitting and training a various range of predictive
    regression models and all their related methods and attributes.

    Attributes
    ----------
    algorithm : str
        sklearn regression algorithm to build and fit model with.
    parameters : dict (default = {})
        parameters to use for building regression model, by default the models'
        default parameters are used.
    test_split : float (default = 0.2)
        size of the test data.

    Methods
    -------
    get_model():

    train_test_split(X, Y, scale = True, test_size = 0.2,random_state=None, shuffle=True):

    fit():

    predict():

    save(save_folder):

    copy():

    hyperparameter_tuning(parameters, metric='r2', cv=5):

    model_fitted():

    """
    def __init__(self, algorithm,parameters={}):

        self.algorithm = algorithm
        self.parameters = parameters
        self.test_split = None

        #list of valid models available to use for this class
        self.valid_models = ['PlsRegression','RandomForestRegressor','AdaBoostRegressor',\
                            'BaggingRegressor','DecisionTreeRegressor','LinearRegression',\
                            'Lasso','SVR','KNeighborsRegressor', 'KNN']

        #get closest match of valid model from the input algorithm parameter value
        model_matches = get_close_matches(self.algorithm.lower(),[item.lower() \
            for item in self.valid_models], cutoff=0.4)

        #if algorithm is a valid model then set it to self.algorithm, else raise error
        if model_matches!=[]:
            self.algorithm = model_matches[0]
        else:
            raise ValueError('Input algorithm {} not found in available valid \
                models {}.'.format(self.algorithm, self.valid_models))

        #create instance of algorithm object
        self.model = self.get_model()
        self.model_fit = None

    def get_model(self):
        """
        Create instance of model type specifed by input 'algorithm' argument. If
        parameters input 'parameters' = {} then default parameters are used else set
        the parameters of the model to the values specified in the 'parameters' input
        parameter.

        Returns
        -------
        model : sklearn.model
            instantiated regression model with default or user-specified parameters.
        """
        if self.algorithm.lower() == 'plsregression':

            #get parameters of sklearn model and check that user inputted
            #  parameters are available in the model, only use those that are valid.
            model_params = set(dir(PLSRegression()))
            parameters = [i for i in model_params if i in self.parameters]

            #use default model parameters if user input parameters is empty, else
            #   use user-specified parameters.
            if parameters != {}:
                model = PLSRegression(**self.parameters)
            else:
                model = PLSRegression()

        elif self.algorithm.lower() == 'randomforestregressor':

            model_params = set(dir(RandomForestRegressor()))
            parameters = [i for i in model_params if i in self.parameters]

            if parameters != {}:
                model = RandomForestRegressor(**self.parameters)
            else:
                model = RandomForestRegressor()

        elif self.algorithm.lower() == 'adaboostregressor':

            model_params = set(dir(AdaBoostRegressor()))
            parameters = [i for i in model_params if i in self.parameters]

            if parameters != {}:
                model = AdaBoostRegressor(**self.parameters)
            else:
                model = AdaBoostRegressor()

        elif self.algorithm.lower() == 'baggingregressor':

            model_params = set(dir(BaggingRegressor()))
            parameters = [i for i in model_params if i in self.parameters]

            if parameters != {}:
                model = BaggingRegressor(**self.parameters)
            else:
                model = BaggingRegressor()

        elif self.algorithm.lower() == 'decisiontreeregressor':

            model_params = set(dir(DecisionTreeRegressor()))
            parameters = [i for i in model_params if i in self.parameters]

            if parameters != {}:
                model = DecisionTreeRegressor(**self.parameters)
            else:
                model = DecisionTreeRegressor()

        elif self.algorithm.lower() == 'linearregression':

            model_params = set(dir(LinearRegression()))
            parameters = [i for i in model_params if i in self.parameters]

            if parameters != {}:
                model = LinearRegression(**self.parameters)
            else:
                model = LinearRegression()

        elif self.algorithm.lower() == 'lasso':

            model_params = set(dir(Lasso()))
            parameters = [i for i in model_params if i in self.parameters]

            if parameters != {}:
                model = Lasso(**self.parameters)
            else:
                model = Lasso()

        elif self.algorithm.lower() == 'svr':

            model_params = set(dir(SVR()))
            parameters = [i for i in model_params if i in self.parameters]

            if parameters != {}:
                model = SVR(**self.parameters)
            else:
                model = SVR()

        elif self.algorithm.lower() == 'knn' or \
           self.algorithm.lower() == 'kneighborsregressor':

            model_params = set(dir(KNeighborsRegressor()))
            parameters = [i for i in model_params if i in self.parameters]

            if parameters != {}:
                model = KNeighborsRegressor(**self.parameters)
            else:
                model = KNeighborsRegressor()
        else:
            raise ValueError('Input Algorithm ({}) not found in available valid \
                models'.format(self.algorithm, self.valid_models))

        return model

    def train_test_split(self, X, Y, scale = True, test_size = 0.2,
        random_state=None, shuffle=True):
        """
        Split the X and Y input features and labels into random train and test
        subsets. By default a 80:20 split will be used, whereby 80% of the data
        will be used for training and 20% for testing. By default the input will
        be scaled first such that the mean is removed and features scaled to unit
        variance. By default data is shuffled before the split and random is None.

        Parameters
        ----------
        X : np.ndarray
            array of feaure data.
        Y : np.ndarray
            array of observed label values.
        scale : bool (default = True)
            if true then scale the features such that they are standardised.
        test_size : float (default = 0.2)
            proportion of the total dataset to use for testing.
        random_state : float (default = None)
            Controls the shuffling applied to the data before applying the split.
            Popular integer random seeds are 0 and 42.
        shuffle : bool (default = True)
            Whether or not to shuffle the data before splitting.

        Returns
        -------
        self.X_train, self.X_test, self.Y_train, self.Y_test : np.ndarray
            splitted training and test data features and labels.
        """
        #validate that X and Y arrays are of the same size
        if (len(X)!=len(Y)):
            raise ValueError('X and Y input parameters must be of the same length.')

        #reshape input arrays to 2D arrays
        if (X.ndim!=2):
            X = np.reshape(X, (-1,1))
        if (Y.ndim!=2):
            Y = np.reshape(Y, (-1,1))

        #if invalid test size input then set to default 0.2
        if (test_size <= 0 or test_size >=1):
            test_size = 0.2

        self.test_split = test_size     #setting test_size attribute

        #scale X
        if scale:
            X = StandardScaler().fit_transform(X)

        #split X and Y into training and test data
        X_train, X_test, Y_train, Y_test = train_test_split(X, Y,
            test_size=test_size, random_state=random_state, shuffle=shuffle)

        #set X and Y attributes
        self.X_train = X_train
        self.X_test = X_test
        self.Y_train = np.reshape(Y_train, (len(Y_train),))
        self.Y_test = np.reshape(Y_test, (len(Y_test),))

        return self.X_train, self.X_test, self.Y_train, self.Y_test

    def fit(self):
        """
        Fit model to training data and target values X and Y.

        Returns
        -------
        self.model_fit : np.ndarray
            fitted sklearn model of type specified by algorithm attribute.
        """
        self.model_fit = self.model.fit(self.X_train, self.Y_train)
        return self.model_fit

    def predict(self):
        """
        Predict the target values of unseen test data using the model.

        Returns
        -------
        self.model_fit.predict(self.X_test) : np.ndarray
            array of predicted target values for unseen test data.
        """
        return self.model_fit.predict(self.X_test)

    def save(self, save_folder):
        """
        Save fitted model to specified save_folder.

        Parameters
        ----------
        save_folder : str
            folder to save model to.
        """
        save_path = os.path.join(save_folder, 'model.pkl')

        try:
            with open(save_path, 'wb') as file:
                pickle.dump(self.model, file)
        except (IOError, OSError, pickle.PickleError, pickle.UnpicklingError):
            print("Error pickling model with path: {} ".format(save_path))

    def hyperparameter_tuning(self, parameters, metric='r2', cv=5):
        """
        Hyperparamter tuning of model to find its optimal arrangment of parameters
        using a Grid Search.

        Parameters
        ----------
        parameters : dict
            dictionary of parameter names and their values.
        metric : str (default = r2)
            scoring metric used to evaluate the performance of the cross-validated
            model on the test set.
        cv : int (default = 5)
            Determines the cross-validation splitting strategy.

        Returns
        -------
        metrics_df: pd.DataFrame
            dataframe of best results and the associated parameters from the
            hyperparameter tuning process.
        """
        #input parameters parameter must be a dict, if not raise error
        if not isinstance(parameters, dict):
            raise TypeError('Parameters argument must be of type dict.')

        #input metric must be in available scoring metrics, if not raise error
        if metric not in sorted(SCORERS.keys()):
            raise UndefinedMetricWarning('Invalid scoring metric, \
                {} not in available Sklearn Scoring Metrics: {}\n'.format(
                    metric, SCORERS.keys()))

        #cv must be of type int and be between 5 and 10, if not then default of 5 is used
        if not (isinstance(cv, int)) or (cv<5 or cv>10):
            cv = 5

        #iterate through all parameter names to check if they are correct for model
        #   if parameter not found in model params then delete from dictionary
        for p in list(parameters.keys()):
            if p not in (list(self.model.get_params().keys())):
                del parameters[p]

        #grid search of hyperparameter space for model
        model_copy = self.copy()
        grid_search = GridSearchCV(estimator=model_copy, param_grid=parameters,\
            n_jobs=-1, cv=cv, scoring=metric,error_score=0)

        #fit X and Y to best model found in grid search
        grid_result = grid_search.fit(self.X_train, self.Y_train)

        #get best grid search metics values
        mean_test = grid_result.cv_results_['mean_test_score']
        std_test = grid_result.cv_results_['std_test_score']
        params = grid_result.cv_results_['params']

        #predict values of unseen test data
        best_model_pred = grid_result.predict(self.X_test)

        #create instance of Evaluate class and calculate metrics from best model
        eval = Evaluate(self.Y_test,best_model_pred)

        #print out results of grid search
        print('\n#############################################################')
        print('#################### Hyperparamter Results ####################')
        print('#############################################################\n')

        print('######################### Parameters ########################\n')
        print('# Best Params -> {}'.format(grid_result.best_params_))
        print('# Model Type -> {}'.format(repr(self)))
        print('# Scoring Metric -> {}'.format(metric))
        print('# CV -> {}'.format(cv))
        print('# Test Split -> {}'.format(self.test_size))

        print('########################## Metrics ###########################\n')
        print('# Best Score -> {}'.format(grid_result.best_score_))
        print('# RMSE: {} '.format(eval.rmse))
        print('# MSE: {} '.format(eval.mse))
        print('# MAE: {}'.format(eval.mae))
        print('# RPD {}'.format(eval.rpd))
        print('# Variance {}\n'.format(eval.explained_var))
        print('###############################################################')

    def copy(self):
        """
        Make a copy of the sklearn model stored in self.model instance variable.

        Returns
        -------
        model_copy : sklearn.model
            deep copy of model.
        """
        model_copy = self.model
        return model_copy

    def model_fitted(self):
        """
        Return if model has been fitted, true or false.

        Returns
        -------
        True/False : bool
            true if model (self.model) has been fitted, false if not.
        """
        return (self.model_fit != None)

######################          Getters & Setters          ######################

    @property
    def test_split(self):
        return self._test_split

    @test_split.setter
    def test_split(self, val):
        self._test_split = val

    @property
    def valid_models(self):
        return self._valid_models

    @valid_models.setter
    def valid_models(self,val):
        self._valid_models = val

    @property
    def parameters(self):
        return self._parameters

    @parameters.setter
    def parameters(self,val):
        self._parameters = val

    @property
    def algorithm(self):
        return self._algorithm

    @algorithm.setter
    def algorithm(self,val):
        self._algorithm = val

    @property
    def model_fit(self):
        return self._model_fit

    @model_fit.setter
    def model_fit(self,val):
        self._model_fit = val

    @property
    def valid_models(self):
        return self._valid_models

    @valid_models.setter
    def valid_models(self,val):
        self._valid_models = val

################################################################################

    def __str__(self):
        return "Model of type {} using {} parameters, model fit = {}".format(
            type(self.model).__name__, self.parameters, self.modelFitted())

    def __repr__(self):
        return type(self.model).__name__

    def __sizeof__(self):
        """ Get size of sklearn model """
        return self.model.__sizeof__()



# -*- coding: utf-8 -*-
#  Copyright (c) 2016-2017, Zhijiang Yao, Jie Dong and Dongsheng Cao
#  All rights reserved.
#  This file is part of the PyBioMed.
#  The contents are covered by the terms of the BSD license
#  which is included in the file license.txt, found at the root
#  of the PyBioMed source tree.
"""
#####################################################################################
This module is used for checking whether the input protein sequence is valid amino acid

sequence. You can freely use and distribute it. If you hava any problem, you could

contact with us timely!

Authors: Zhijiang Yao and Dongsheng Cao.

Date: 2016.06.04

Email: gadsby@163.com

#####################################################################################

"""

AALetter = [
    "A",
    "R",
    "N",
    "D",
    "C",
    "E",
    "Q",
    "G",
    "H",
    "I",
    "L",
    "K",
    "M",
    "F",
    "P",
    "S",
    "T",
    "W",
    "Y",
    "V",
]


def ProteinCheck(ProteinSequence):
    """
    ###################################################################################
    Check whether the protein sequence is a valid amino acid sequence or not

    Usage:

    result=ProteinCheck(protein)

    Input: protein is a pure protein sequence.

    Output: if the check is no problem, result will return the length of protein.

    if the check has problems, result will return 0.
    ###################################################################################
    """

    NumPro = len(ProteinSequence)
    for i in ProteinSequence:
        if i not in AALetter:
            flag = 0
            break
        else:
            flag = NumPro

    return flag


if __name__ == "__main__":
    protein = "ADGCGVGEGTGQGPMCNCMCMKWVYADEDAADLESDSFADEDASLESDSFPWSNQRVFCSFADEDASU"
    print(ProteinCheck(protein))
